"""
The following test suit serves as a set of examples of how to integrate different classes
into our AST and use them.
"""
# stdlib
from functools import partial
from importlib import reload
import sys
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union as TypeUnion

# third party
import pytest

# syft absolute
import syft
from syft.ast.globals import Globals
from syft.core.node.abstract.node import AbstractNodeClient
from syft.core.node.common.client import Client
from syft.lib import lib_ast

# syft relative
from . import module_test

sys.modules["module_test"] = module_test

module_test_methods = [
    ("module_test.A", "module_test.A"),
    ("module_test.A.__len__", "syft.lib.python.Int"),
    ("module_test.A.__iter__", "syft.lib.python.Iterator"),
    ("module_test.A.__next__", "syft.lib.python.Int"),
    ("module_test.A.test_method", "syft.lib.python.Int"),
    ("module_test.A.test_property", "syft.lib.python.Float"),
    ("module_test.A._private_attr", "syft.lib.python.Float"),
    ("module_test.A.static_method", "syft.lib.python.Float"),
    ("module_test.A.static_attr", "syft.lib.python.Int"),
    ("module_test.B.Car", "module_test.B"),
    ("module_test.C", "module_test.C"),
    ("module_test.C.type_reload_func", "syft.lib.python._SyNone"),
    ("module_test.C.obj_reload_func", "syft.lib.python._SyNone"),
    ("module_test.C.dummy_reloadable_func", "syft.lib.python.Int"),
    ("module_test.global_value", "syft.lib.python.Int"),
    ("module_test.global_function", "syft.lib.python.Int"),
]


def update_ast_test(
    ast_or_client: TypeUnion[Globals, AbstractNodeClient],
    methods: List[Tuple[str, str]],
) -> None:
    if isinstance(ast_or_client, Globals):
        ast = ast_or_client
        test_ast = create_ast_test(client=None, methods=methods)
        ast.add_attr(attr_name="module_test", attr=test_ast.attrs["module_test"])
    elif isinstance(ast_or_client, AbstractNodeClient):
        client = ast_or_client
        test_ast = create_ast_test(client=client, methods=methods)
        client.lib_ast.attrs["module_test"] = test_ast.attrs["module_test"]
        setattr(client, "module_test", test_ast.attrs["module_test"])
    else:
        raise ValueError(
            f"Expected param of type (Globals, AbstractNodeClient), but got {type(ast_or_client)}"
        )


def create_ast_test(
    client: Optional[AbstractNodeClient], methods: List[Tuple[str, str]]
) -> Globals:
    ast = Globals(client)

    for method, return_type in methods:
        ast.add_path(
            path=method, framework_reference=module_test, return_type_name=return_type
        )

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast


@pytest.fixture(autouse=True, scope="module")
def register_module_test() -> None:
    # Make lib_ast contain the specific methods/attributes
    update_ast_test(ast_or_client=syft.lib_ast, methods=module_test_methods)

    # Make sure that when we register a new client it would update the specific AST
    lib_ast.loaded_lib_constructors["module_test"] = partial(
        update_ast_test, methods=module_test_methods
    )


@pytest.fixture()
def custom_client() -> Client:
    alice = syft.VirtualMachine(name="alice")
    alice_client = alice.get_root_client()

    return alice_client


def test_len(custom_client: Client) -> None:
    a_ptr = custom_client.module_test.A()
    result_from_ptr = a_ptr.__len__()

    a = module_test.A()
    result = len(a)

    assert result == result_from_ptr


def test_iter(custom_client: Client) -> None:
    a_ptr = custom_client.module_test.A()
    iter_from_ptr = a_ptr.__iter__()
    a = module_test.A()
    iter_from_obj = iter(a)

    for _ in range(1, module_test.A.static_attr):
        assert next(iter_from_ptr).get() == next(iter_from_obj)


def test_method(custom_client: Client) -> None:
    a_ptr = custom_client.module_test.A()
    result_ptr = a_ptr.test_method()

    a = module_test.A()
    result = a.test_method()

    assert result == result_ptr.get()


def test_property_get(custom_client: Client) -> None:
    a_ptr = custom_client.module_test.A()
    result_ptr = a_ptr.test_property

    a = module_test.A()
    result = a.test_property

    assert result == result_ptr.get()


def test_property_set(custom_client: Client) -> None:
    value_to_set = 7.5

    a_ptr = custom_client.module_test.A()
    a_ptr.test_property = value_to_set
    result_ptr = a_ptr.test_property

    a = module_test.A()
    a.test_property = value_to_set
    result = a.test_property

    assert result == result_ptr.get()  # type: ignore


def test_slot_get(custom_client: Client) -> None:
    a_ptr = custom_client.module_test.A()
    result_ptr = a_ptr._private_attr

    a = module_test.A()
    result = a._private_attr

    assert result == result_ptr.get()


def test_slot_set(custom_client: Client) -> None:
    value_to_set = 7.5

    a_ptr = custom_client.module_test.A()
    a_ptr._private_attr = value_to_set
    result_ptr = a_ptr._private_attr

    a = module_test.A()
    a._private_attr = value_to_set
    result = a._private_attr

    assert result == result_ptr.get()  # type: ignore


def test_global_function(custom_client: Client) -> None:
    result_ptr = custom_client.module_test.global_function()
    result = module_test.global_function()

    assert result == result_ptr.get()


def test_global_attribute_get(custom_client: Client) -> None:
    result_ptr = custom_client.module_test.global_value
    result = module_test.global_value

    assert result == result_ptr.get()


def test_global_attribute_set(custom_client: Client) -> None:
    global module_test

    set_value = 5

    custom_client.module_test.global_value = set_value
    result_ptr = custom_client.module_test.global_value
    sy_result = result_ptr.get()  # type: ignore

    module_test = reload(module_test)
    module_test.global_value = set_value
    local_result = module_test.global_value

    assert local_result == sy_result


def test_static_method(custom_client: Client) -> None:
    result_ptr = custom_client.module_test.A.static_method()
    result = module_test.A.static_method()
    assert result == result_ptr.get()


def test_static_attribute_get(custom_client: Client) -> None:
    result_ptr = custom_client.module_test.A.static_attr
    result = module_test.A.static_attr

    assert result == result_ptr.get()


def test_static_attribute_set(custom_client: Client) -> None:
    value_to_set = 5

    custom_client.module_test.A.static_attr = value_to_set
    result_ptr = custom_client.module_test.A.static_attr

    module_test.A.static_attr = value_to_set
    result = module_test.A.static_attr

    assert result == result_ptr.get()  # type: ignore


def test_enum(custom_client: Client) -> None:
    result_ptr = custom_client.module_test.B.Car
    result = module_test.B.Car

    assert result == result_ptr.get()


def test_dynamic_ast_type(custom_client: Client) -> None:
    custom_client.module_test.C.type_reload_func()
    obj_ptr = custom_client.module_test.C()
    result_ptr = obj_ptr.dummy_reloadable_func()

    assert result_ptr.get() == 1


def test_dynamic_ast_obj(custom_client: Client) -> None:
    obj_ptr = custom_client.module_test.C()
    obj_ptr.obj_reload_func()
    result_ptr = obj_ptr.dummy_reloadable_func()

    assert result_ptr.get() == 2



# stdlib
from typing import Any
from typing import Callable

# syft relative
from ...messages.role_messages import CreateRoleMessage
from ...messages.role_messages import DeleteRoleMessage
from ...messages.role_messages import GetRoleMessage
from ...messages.role_messages import GetRolesMessage
from ...messages.role_messages import UpdateRoleMessage
from .request_api import GridRequestAPI


class RoleRequestAPI(GridRequestAPI):
    response_key: str = "role"

    def __init__(self, send: Callable):
        super().__init__(
            create_msg=CreateRoleMessage,
            get_msg=GetRoleMessage,
            get_all_msg=GetRolesMessage,
            update_msg=UpdateRoleMessage,
            delete_msg=DeleteRoleMessage,
            send=send,
            response_key=RoleRequestAPI.response_key,
        )

    def __getitem__(self, key: int) -> Any:
        return self.get(role_id=key)

    def __delitem__(self, key: int) -> None:
        self.delete(role_id=key)



# stdlib
from typing import Any

# third party
import pytest

# syft absolute
import syft as sy


def get_permission(
    obj: Any, node: sy.VirtualMachine, client: sy.VirtualMachineClient
) -> None:
    remote_obj = node.store[obj.id_at_location]
    remote_obj.read_permissions[client.verify_key] = obj.id_at_location


inputs = {
    "__and__": [],
    "__eq__": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "__ge__": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "__gt__": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "__iand__": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "__ior__": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "__isub__": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "__ixor__": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "__le__": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "__len__": [[]],
    "__lt__": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "__ne__": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "__or__": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "__sub__": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "__xor__": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "add": [[10], [24], [42], [-1]],
    "clear": [[]],
    "difference": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "difference_update": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "discard": [[10], [24], [42], [-1]],
    "intersection": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "intersection_update": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "isdisjoint": [{24}],
    "issuperset": [{24}],
    "pop": [[]],
    "remove": [[10], [24], [42], [-1]],
    "symmetric_difference_update": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "symmetric_difference": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "union": [[{42, 24}], [set(range(10))], [set(range(25))]],
    "update": [[{42, 24}], [set(range(10))], [set(range(25))]],
}

objects = [[24], [42], [24, 42], list(range(10))]


@pytest.mark.parametrize("test_object", objects)
@pytest.mark.parametrize("func", inputs.keys())
def test_pointer_objectives(
    test_object, func, node: sy.VirtualMachine, client: sy.VirtualMachineClient
):
    py_obj, sy_obj, remote_sy_obj = (
        set(test_object),
        sy.lib.python.Set(test_object),
        client.syft.lib.python.Set(test_object),
    )
    possible_inputs = inputs[func]

    if not hasattr(py_obj, func):
        return

    py_method = getattr(py_obj, func)
    sy_method = getattr(sy_obj, func)

    if func == "__len__":
        func = "len"

    remote_sy_method = getattr(remote_sy_obj, func)

    for possible_input in possible_inputs:
        try:
            py_res = py_method(*possible_input)
        except Exception as py_e:
            py_res = str(py_e)

        try:
            sy_res = sy_method(*possible_input)
        except Exception as sy_e:
            sy_res = str(sy_e)

        try:
            remote_sy_res = remote_sy_method(*possible_input)
            get_permission(remote_sy_res, node, client)
            remote_sy_res = remote_sy_res.get()
        except Exception as remote_sy_e:
            remote_sy_res = str(remote_sy_e)

        if isinstance(py_res, float):
            py_res = int(py_res * 1000) / 1000
            sy_res = int(sy_res * 1000) / 1000
            remote_sy_res = int(remote_sy_res * 1000) / 1000

        assert py_res == sy_res
        assert sy_res == remote_sy_res



# stdlib
from typing import Callable
from typing import Optional

# third party
from nacl.signing import VerifyKey

# syft relative
from .....logger import debug
from .....logger import traceback_and_raise
from ....common.message import SyftMessage
from ...abstract.node import AbstractNode


class AuthorizationException(Exception):
    pass


def service_auth(
    root_only: bool = False,
    admin_only: bool = False,
    cpl_ofcr_only: bool = False,
    existing_users_only: bool = False,
    guests_welcome: bool = False,
    register_new_guests: bool = False,
) -> Callable:
    def decorator(func: Callable) -> Callable:
        def process(
            node: AbstractNode, msg: SyftMessage, verify_key: VerifyKey
        ) -> Optional[SyftMessage]:
            debug(f"> Checking {msg.pprint} 🔑 Matches {node.pprint} root 🗝")

            if root_only:
                keys = (
                    f"> Matching 🔑 {node.key_emoji(key=verify_key)}  == "
                    + f"{node.key_emoji(key=node.root_verify_key)}  🗝"
                )
                debug(keys)
                if verify_key != node.root_verify_key:
                    debug(f"> ❌ Auth FAILED {msg.pprint}")
                    traceback_and_raise(
                        AuthorizationException(
                            "You are not Authorized to access this service"
                        )
                    )
                else:
                    debug(f"> ✅ Auth Succeeded {msg.pprint} 🔑 == 🗝")

            elif admin_only:
                if (
                    verify_key not in node.admin_verify_key_registry
                    and verify_key != node.root_verify_key
                ):
                    debug(f"> ❌ Auth FAILED {msg.pprint}")
                    traceback_and_raise(
                        AuthorizationException("User lacks Administrator credentials.")
                    )

            elif cpl_ofcr_only:
                if (
                    verify_key not in node.cpl_ofcr_verify_key_registry
                    and verify_key != node.root_verify_key
                ):
                    debug(f"> ❌ Auth FAILED {msg.pprint}")
                    traceback_and_raise(
                        AuthorizationException(
                            "User lacks Compliance Officer credentials."
                        )
                    )

            elif existing_users_only:
                if verify_key not in node.guest_verify_key_registry:
                    debug(f"> ❌ Auth FAILED {msg.pprint}")
                    traceback_and_raise(AuthorizationException("User not known."))

            elif guests_welcome:
                if register_new_guests:
                    node.guest_verify_key_registry.add(verify_key)

            else:
                traceback_and_raise(
                    Exception("You must configure services auth with a flag.")
                )

            # Can be None because not all functions reply
            return func(node=node, msg=msg, verify_key=verify_key)

        return process

    return decorator



# flake8: noqa
"""
Tests copied from cpython test suite:
https://github.com/python/cpython/blob/3.9/Lib/test/test_complex.py
"""

# stdlib
from math import atan2
from math import copysign
from math import isnan
import operator
from random import random
import sys
import unittest

# third party
import pytest

# syft absolute
from syft.lib.python.complex import Complex

INF = float("inf")
NAN = float("nan")

VALID_UNDERSCORE_LITERALS = [
    "0_0_0",
    "4_2",
    "1_0000_0000",
    "0b1001_0100",
    "0xffff_ffff",
    "0o5_7_7",
    "1_00_00.5",
    "1_00_00.5e5",
    "1_00_00e5_1",
    "1e1_0",
    ".1_4",
    ".1_4e1",
    "0b_0",
    "0x_f",
    "0o_5",
    "1_00_00j",
    "1_00_00.5j",
    "1_00_00e5_1j",
    ".1_4j",
    "(1_2.5+3_3j)",
    "(.5_6j)",
]
INVALID_UNDERSCORE_LITERALS = [
    # Trailing underscores:
    "0_",
    "42_",
    "1.4j_",
    "0x_",
    "0b1_",
    "0xf_",
    "0o5_",
    "0 if 1_Else 1",
    # Underscores in the base selector:
    "0_b0",
    "0_xf",
    "0_o5",
    # Old-style octal, still disallowed:
    "0_7",
    "09_99",
    # Multiple consecutive underscores:
    "4_______2",
    "0.1__4",
    "0.1__4j",
    "0b1001__0100",
    "0xffff__ffff",
    "0x___",
    "0o5__77",
    "1e1__0",
    "1e1__0j",
    # Underscore right before a dot:
    "1_.4",
    "1_.4j",
    # Underscore right after a dot:
    "1._4",
    "1._4j",
    "._5",
    "._5j",
    # Underscore right after a sign:
    "1.0e+_1",
    "1.0e+_1j",
    # Underscore right before j:
    "1.4_j",
    "1.4e5_j",
    # Underscore right before e:
    "1_e1",
    "1.4_e1",
    "1.4_e1j",
    # Underscore right after e:
    "1e_1",
    "1.4e_1",
    "1.4e_1j",
    # Complex cases with parens:
    "(1+1.5_j_)",
    "(1+1.5_j)",
]


class ComplexTest(unittest.TestCase):
    @staticmethod
    def assertIs(a, b):
        assert a == b

    def assertAlmostEqual(self, a, b):
        if isinstance(a, complex):
            if isinstance(b, complex):
                unittest.TestCase.assertAlmostEqual(self, a.real, b.real)
                unittest.TestCase.assertAlmostEqual(self, a.imag, b.imag)
            else:
                unittest.TestCase.assertAlmostEqual(self, a.real, b)
                unittest.TestCase.assertAlmostEqual(self, a.imag, 0.0)
        else:
            if isinstance(b, complex):
                unittest.TestCase.assertAlmostEqual(self, a, b.real)
                unittest.TestCase.assertAlmostEqual(self, 0.0, b.imag)
            else:
                unittest.TestCase.assertAlmostEqual(self, a, b)

    def assertCloseAbs(self, x, y, eps=1e-9):
        """Return true iff floats x and y "are close"."""
        # put the one with larger magnitude second
        if abs(x) > abs(y):
            x, y = y, x
        if y == 0:
            return abs(x) < eps
        if x == 0:
            return abs(y) < eps
        # check that relative difference < eps
        self.assertTrue(abs((x - y) / y) < eps)

    def assertFloatsAreIdentical(self, x, y):
        """assert that floats x and y are identical, in the sense that:
        (1) both x and y are nans, or
        (2) both x and y are infinities, with the same sign, or
        (3) both x and y are zeros, with the same sign, or
        (4) x and y are both finite and nonzero, and x == y

        """
        msg = "floats {!r} and {!r} are not identical"

        if isnan(x) or isnan(y):
            if isnan(x) and isnan(y):
                return
        elif x == y:
            if x != 0.0:
                return
            # both zero; check that signs match
            elif copysign(1.0, x) == copysign(1.0, y):
                return
            else:
                msg += ": zeros have different signs"
        self.fail(msg.format(x, y))

    def assertClose(self, x, y, eps=1e-9):
        """Return true iff complexes x and y "are close"."""
        self.assertCloseAbs(x.real, y.real, eps)
        self.assertCloseAbs(x.imag, y.imag, eps)

    def check_div(self, x, y):
        """Compute Complex z=x*y, and check that z/x==y and z/y==x."""
        z = x * y
        if x != 0:
            q = z / x
            self.assertClose(q, y)
            q = z.__truediv__(x)
            self.assertClose(q, y)
        if y != 0:
            q = z / y
            self.assertClose(q, x)
            q = z.__truediv__(y)
            self.assertClose(q, x)

    @pytest.mark.slow
    def test_truediv(self):
        simple_real = [float(i) for i in range(-5, 6)]
        simple_complex = [Complex(x, y) for x in simple_real for y in simple_real]
        for x in simple_complex:
            for y in simple_complex:
                self.check_div(x, y)

        # A naive Complex division algorithm (such as in 2.0) is very prone to
        # nonsense errors for these (overflows and underflows).
        self.check_div(Complex(1e200, 1e200), 1 + 0j)
        self.check_div(Complex(1e-200, 1e-200), 1 + 0j)

        # Just for fun.
        for i in range(100):
            self.check_div(Complex(random(), random()), Complex(random(), random()))

        self.assertRaises(ZeroDivisionError, Complex.__truediv__, 1 + 1j, 0 + 0j)
        self.assertRaises(OverflowError, pow, 1e200 + 1j, 1e200 + 1j)

        self.assertAlmostEqual(Complex.__truediv__(2 + 0j, 1 + 1j), 1 - 1j)
        self.assertRaises(ZeroDivisionError, Complex.__truediv__, 1 + 1j, 0 + 0j)

        for denom_real, denom_imag in [(0, NAN), (NAN, 0), (NAN, NAN)]:
            z = Complex(0, 0) / Complex(denom_real, denom_imag)
            self.assertTrue(isnan(z.real))
            self.assertTrue(isnan(z.imag))

    def test_floordiv(self):
        self.assertRaises(TypeError, Complex.__floordiv__, 3 + 0j, 1.5 + 0j)
        self.assertRaises(TypeError, Complex.__floordiv__, 3 + 0j, 0 + 0j)

    def test_richcompare(self):
        self.assertIs(Complex.__eq__(1 + 1j, 1 << 10000), False)
        self.assertIs(Complex.__lt__(1 + 1j, None), NotImplemented)
        self.assertIs(Complex.__eq__(1 + 1j, 1 + 1j), True)
        self.assertIs(Complex.__eq__(1 + 1j, 2 + 2j), False)
        self.assertIs(Complex.__ne__(1 + 1j, 1 + 1j), False)
        self.assertIs(Complex.__ne__(1 + 1j, 2 + 2j), True)
        for i in range(1, 100):
            f = i / 100.0
            self.assertIs(Complex.__eq__(f + 0j, f), True)
            self.assertIs(Complex.__ne__(f + 0j, f), False)
            self.assertIs(Complex.__eq__(Complex(f, f), f), False)
            self.assertIs(Complex.__ne__(Complex(f, f), f), True)
        self.assertIs(Complex.__lt__(1 + 1j, 2 + 2j), NotImplemented)
        self.assertIs(Complex.__le__(1 + 1j, 2 + 2j), NotImplemented)
        self.assertIs(Complex.__gt__(1 + 1j, 2 + 2j), NotImplemented)
        self.assertIs(Complex.__ge__(1 + 1j, 2 + 2j), NotImplemented)
        self.assertRaises(TypeError, operator.lt, 1 + 1j, 2 + 2j)
        self.assertRaises(TypeError, operator.le, 1 + 1j, 2 + 2j)
        self.assertRaises(TypeError, operator.gt, 1 + 1j, 2 + 2j)
        self.assertRaises(TypeError, operator.ge, 1 + 1j, 2 + 2j)
        self.assertIs(operator.eq(1 + 1j, 1 + 1j), True)
        self.assertIs(operator.eq(1 + 1j, 2 + 2j), False)
        self.assertIs(operator.ne(1 + 1j, 1 + 1j), False)
        self.assertIs(operator.ne(1 + 1j, 2 + 2j), True)

    @pytest.mark.slow
    def test_richcompare_boundaries(self):
        def check(n, deltas, is_equal, imag=0.0):
            for delta in deltas:
                i = n + delta
                z = Complex(i, imag)
                self.assertIs(Complex.__eq__(z, i), is_equal(delta))
                self.assertIs(Complex.__ne__(z, i), not is_equal(delta))

        # For IEEE-754 doubles the following should hold:
        #    x in [2 ** (52 + i), 2 ** (53 + i + 1)] -> x mod 2 ** i == 0
        # where the interval is representable, of course.
        for i in range(1, 10):
            pow = 52 + i
            mult = 2 ** i
            check(2 ** pow, range(1, 101), lambda delta: delta % mult == 0)
            check(2 ** pow, range(1, 101), lambda delta: False, float(i))
        check(2 ** 53, range(-100, 0), lambda delta: True)

    def test_mod(self):
        # % is no longer supported on Complex numbers
        self.assertRaises(TypeError, (1 + 1j).__mod__, 0 + 0j)
        self.assertRaises(TypeError, lambda: (3.33 + 4.43j) % 0)
        self.assertRaises(TypeError, (1 + 1j).__mod__, 4.3j)

    def test_divmod(self):
        self.assertRaises(TypeError, divmod, 1 + 1j, 1 + 0j)
        self.assertRaises(TypeError, divmod, 1 + 1j, 0 + 0j)

    def test_pow(self):
        self.assertAlmostEqual(pow(1 + 1j, 0 + 0j), 1.0)
        self.assertAlmostEqual(pow(0 + 0j, 2 + 0j), 0.0)
        self.assertRaises(ZeroDivisionError, pow, 0 + 0j, 1j)
        self.assertAlmostEqual(pow(1j, -1), 1 / 1j)
        self.assertAlmostEqual(pow(1j, 200), 1)
        self.assertRaises(ValueError, pow, 1 + 1j, 1 + 1j, 1 + 1j)

        a = 3.33 + 4.43j
        self.assertEqual(a ** 0j, 1)
        self.assertEqual(a ** 0.0 + 0.0j, 1)

        self.assertEqual(3j ** 0j, 1)
        self.assertEqual(3j ** 0, 1)

        try:
            0j ** a
        except ZeroDivisionError:
            pass
        else:
            self.fail("should fail 0.0 to negative or Complex power")

        try:
            0j ** (3 - 2j)
        except ZeroDivisionError:
            pass
        else:
            self.fail("should fail 0.0 to negative or Complex power")

        # The following is used to exercise certain code paths
        self.assertEqual(a ** 105, a ** 105)
        self.assertEqual(a ** -105, a ** -105)
        self.assertEqual(a ** -30, a ** -30)

        self.assertEqual(0.0j ** 0, 1)

        b = 5.1 + 2.3j
        self.assertRaises(ValueError, pow, a, b, 0)

    def test_boolcontext(self):
        for i in range(100):
            self.assertTrue(Complex(random() + 1e-6, random() + 1e-6))
        self.assertTrue(not Complex(0.0, 0.0))

    def test_conjugate(self):
        self.assertClose(Complex(5.3, 9.8).conjugate(), 5.3 - 9.8j)

    @pytest.mark.slow
    def test_constructor(self):
        class OS:
            def __init__(self, value):
                self.value = value

            def __complex__(self):
                return self.value

        class NS(object):
            def __init__(self, value):
                self.value = value

            def __complex__(self):
                return self.value

        self.assertEqual(Complex(OS(1 + 10j)), 1 + 10j)
        self.assertEqual(Complex(NS(1 + 10j)), 1 + 10j)
        self.assertRaises(TypeError, Complex, OS(None))
        self.assertRaises(TypeError, Complex, NS(None))
        self.assertRaises(TypeError, Complex, {})
        self.assertRaises(TypeError, Complex, NS(1.5))
        self.assertRaises(TypeError, Complex, NS(1))

        self.assertAlmostEqual(Complex("1+10j"), 1 + 10j)
        self.assertAlmostEqual(Complex(10), 10 + 0j)
        self.assertAlmostEqual(Complex(10.0), 10 + 0j)
        self.assertAlmostEqual(Complex(10), 10 + 0j)
        self.assertAlmostEqual(Complex(10 + 0j), 10 + 0j)
        self.assertAlmostEqual(Complex(1, 10), 1 + 10j)
        self.assertAlmostEqual(Complex(1, 10), 1 + 10j)
        self.assertAlmostEqual(Complex(1, 10.0), 1 + 10j)
        self.assertAlmostEqual(Complex(1, 10), 1 + 10j)
        self.assertAlmostEqual(Complex(1, 10), 1 + 10j)
        self.assertAlmostEqual(Complex(1, 10.0), 1 + 10j)
        self.assertAlmostEqual(Complex(1.0, 10), 1 + 10j)
        self.assertAlmostEqual(Complex(1.0, 10), 1 + 10j)
        self.assertAlmostEqual(Complex(1.0, 10.0), 1 + 10j)
        self.assertAlmostEqual(Complex(3.14 + 0j), 3.14 + 0j)
        self.assertAlmostEqual(Complex(3.14), 3.14 + 0j)
        self.assertAlmostEqual(Complex(314), 314.0 + 0j)
        self.assertAlmostEqual(Complex(314), 314.0 + 0j)
        self.assertAlmostEqual(Complex(3.14 + 0j, 0j), 3.14 + 0j)
        self.assertAlmostEqual(Complex(3.14, 0.0), 3.14 + 0j)
        self.assertAlmostEqual(Complex(314, 0), 314.0 + 0j)
        self.assertAlmostEqual(Complex(314, 0), 314.0 + 0j)
        self.assertAlmostEqual(Complex(0j, 3.14j), -3.14 + 0j)
        self.assertAlmostEqual(Complex(0.0, 3.14j), -3.14 + 0j)
        self.assertAlmostEqual(Complex(0j, 3.14), 3.14j)
        self.assertAlmostEqual(Complex(0.0, 3.14), 3.14j)
        self.assertAlmostEqual(Complex("1"), 1 + 0j)
        self.assertAlmostEqual(Complex("1j"), 1j)
        self.assertAlmostEqual(Complex(), 0)
        self.assertAlmostEqual(Complex("-1"), -1)
        self.assertAlmostEqual(Complex("+1"), +1)
        self.assertAlmostEqual(Complex("(1+2j)"), 1 + 2j)
        self.assertAlmostEqual(Complex("(1.3+2.2j)"), 1.3 + 2.2j)
        self.assertAlmostEqual(Complex("3.14+1J"), 3.14 + 1j)
        self.assertAlmostEqual(Complex(" ( +3.14-6J )"), 3.14 - 6j)
        self.assertAlmostEqual(Complex(" ( +3.14-J )"), 3.14 - 1j)
        self.assertAlmostEqual(Complex(" ( +3.14+j )"), 3.14 + 1j)
        self.assertAlmostEqual(Complex("J"), 1j)
        self.assertAlmostEqual(Complex("( j )"), 1j)
        self.assertAlmostEqual(Complex("+J"), 1j)
        self.assertAlmostEqual(Complex("( -j)"), -1j)
        self.assertAlmostEqual(Complex("1e-500"), 0.0 + 0.0j)
        self.assertAlmostEqual(Complex("-1e-500j"), 0.0 - 0.0j)
        self.assertAlmostEqual(Complex("-1e-500+1e-500j"), -0.0 + 0.0j)

        class complex2(complex):
            pass

        self.assertAlmostEqual(Complex(complex2(1 + 1j)), 1 + 1j)
        self.assertAlmostEqual(Complex(real=17, imag=23), 17 + 23j)
        self.assertAlmostEqual(Complex(real=17 + 23j), 17 + 23j)
        self.assertAlmostEqual(Complex(real=17 + 23j, imag=23), 17 + 46j)
        self.assertAlmostEqual(Complex(real=1 + 2j, imag=3 + 4j), -3 + 5j)

        # check that the sign of a zero in the real or imaginary part
        # is preserved when constructing from two floats.  (These checks
        # are harmless on systems without support for signed zeros.)
        def split_zeros(x):
            """Function that produces different results for 0. and -0."""
            return atan2(x, -1.0)

        self.assertEqual(split_zeros(Complex(1.0, 0.0).imag), split_zeros(0.0))
        self.assertEqual(split_zeros(Complex(1.0, -0.0).imag), split_zeros(-0.0))
        self.assertEqual(split_zeros(Complex(0.0, 1.0).real), split_zeros(0.0))
        self.assertEqual(split_zeros(Complex(-0.0, 1.0).real), split_zeros(-0.0))

        # c = 3.14 + 1j
        # self.assertTrue(Complex(c) is c)
        # del c

        self.assertRaises(TypeError, Complex, "1", "1")
        self.assertRaises(TypeError, Complex, 1, "1")

        # SF bug 543840:  Complex(string) accepts strings with \0
        # Fixed in 2.3.
        self.assertRaises(ValueError, Complex, "1+1j\0j")

        self.assertRaises(TypeError, int, 5 + 3j)
        self.assertRaises(TypeError, int, 5 + 3j)
        self.assertRaises(TypeError, float, 5 + 3j)
        self.assertRaises(ValueError, Complex, "")
        # FIXME improve Complex.__new__() to have the same TypeError as complex
        # self.assertRaises(TypeError, Complex, None)
        # self.assertRaisesRegex(TypeError, "not 'NoneType'", Complex, None)
        self.assertRaises(ValueError, Complex, "\0")
        self.assertRaises(ValueError, Complex, "3\09")
        self.assertRaises(TypeError, Complex, "1", "2")
        self.assertRaises(TypeError, Complex, "1", 42)
        self.assertRaises(TypeError, Complex, 1, "2")
        self.assertRaises(ValueError, Complex, "1+")
        self.assertRaises(ValueError, Complex, "1+1j+1j")
        self.assertRaises(ValueError, Complex, "--")
        self.assertRaises(ValueError, Complex, "(1+2j")
        self.assertRaises(ValueError, Complex, "1+2j)")
        self.assertRaises(ValueError, Complex, "1+(2j)")
        self.assertRaises(ValueError, Complex, "(1+2j)123")
        self.assertRaises(ValueError, Complex, "x")
        self.assertRaises(ValueError, Complex, "1j+2")
        self.assertRaises(ValueError, Complex, "1e1ej")
        self.assertRaises(ValueError, Complex, "1e++1ej")
        self.assertRaises(ValueError, Complex, ")1+2j(")
        self.assertRaisesRegex(
            TypeError,
            "first argument must be a string or a number, not 'dict'",
            Complex,
            {1: 2},
            1,
        )
        self.assertRaisesRegex(
            TypeError,
            "second argument must be a number, not 'dict'",
            Complex,
            1,
            {1: 2},
        )
        # the following three are accepted by Python 2.6
        self.assertRaises(ValueError, Complex, "1..1j")
        self.assertRaises(ValueError, Complex, "1.11.1j")
        self.assertRaises(ValueError, Complex, "1e1.1j")

        # check that Complex accepts long unicode strings
        self.assertEqual(type(Complex("1" * 500)), Complex)
        # check whitespace processing
        self.assertEqual(Complex("\N{EM SPACE}(\N{EN SPACE}1+1j ) "), 1 + 1j)
        # Invalid unicode string
        # See bpo-34087
        self.assertRaises(ValueError, Complex, "\u3053\u3093\u306b\u3061\u306f")

        class EvilExc(Exception):
            pass

        class evilcomplex:
            def __complex__(self):
                raise EvilExc

        self.assertRaises(EvilExc, Complex, evilcomplex())

        class float2:
            def __init__(self, value):
                self.value = value

            def __float__(self):
                return self.value

        self.assertAlmostEqual(Complex(float2(42.0)), 42)
        self.assertAlmostEqual(Complex(real=float2(17.0), imag=float2(23.0)), 17 + 23j)
        self.assertRaises(TypeError, Complex, float2(None))

        # FIXME doesn't work
        # class MyIndex:
        #     def __init__(self, value):
        #         self.value = value

        #     def __index__(self):
        #         return self.value

        # self.assertAlmostEqual(Complex(MyIndex(42)), 42.0 + 0.0j)
        # self.assertAlmostEqual(Complex(123, MyIndex(42)), 123.0 + 42.0j)
        # self.assertRaises(OverflowError, Complex, MyIndex(2 ** 2000))
        # self.assertRaises(OverflowError, Complex, 123, MyIndex(2 ** 2000))

        class MyInt:
            def __int__(self):
                return 42

        self.assertRaises(TypeError, Complex, MyInt())
        self.assertRaises(TypeError, Complex, 123, MyInt())

        class complex0(Complex):
            """Test usage of __complex__() when inheriting from 'Complex'"""

            def __complex__(self):
                return 42j

        class complex1(Complex):
            """Test usage of __complex__() with a __new__() method"""

            def __new__(self, value=0j):
                return Complex.__new__(self, 2 * value)

            def __complex__(self):
                return self

        class complex2(Complex):
            """Make sure that __complex__() calls fail if anything other than a
            Complex is returned"""

            def __complex__(self):
                return None

        self.assertEqual(Complex(complex0(1j)), 42j)
        if sys.version_info >= (3, 7):
            # Only deprecated from python 3.7
            with self.assertWarns(DeprecationWarning):
                self.assertEqual(Complex(complex1(1j)), 2j)
        self.assertRaises(TypeError, Complex, complex2(1j))

    # @support.requires_IEEE_754
    def test_constructor_special_numbers(self):
        class complex2(complex):
            pass

        for x in 0.0, -0.0, INF, -INF, NAN:
            for y in 0.0, -0.0, INF, -INF, NAN:
                with self.subTest(x=x, y=y):
                    z = Complex(x, y)
                    self.assertFloatsAreIdentical(z.real, x)
                    self.assertFloatsAreIdentical(z.imag, y)
                    z = complex2(x, y)
                    self.assertIs(type(z), complex2)
                    self.assertFloatsAreIdentical(z.real, x)
                    self.assertFloatsAreIdentical(z.imag, y)
                    z = Complex(complex2(x, y))
                    self.assertIs(type(z), Complex)
                    self.assertFloatsAreIdentical(z.real, x)
                    self.assertFloatsAreIdentical(z.imag, y)
                    z = complex2(Complex(x, y))
                    self.assertIs(type(z), complex2)
                    self.assertFloatsAreIdentical(z.real, x)
                    self.assertFloatsAreIdentical(z.imag, y)

    def test_underscores(self):
        # check underscores
        for lit in VALID_UNDERSCORE_LITERALS:
            if not any(ch in lit for ch in "xXoObB"):
                self.assertEqual(Complex(lit), eval(lit))
                self.assertEqual(Complex(lit), Complex(lit.replace("_", "")))
        for lit in INVALID_UNDERSCORE_LITERALS:
            if lit in ("0_7", "09_99"):  # octals are not recognized here
                continue
            if not any(ch in lit for ch in "xXoObB"):
                self.assertRaises(ValueError, Complex, lit)

    def test_hash(self):
        for x in range(-30, 30):
            self.assertEqual(hash(x), hash(Complex(x, 0)))
            x /= 3.0  # now check against floating point
            self.assertEqual(hash(x), hash(Complex(x, 0.0)))

    def test_abs(self):
        nums = [Complex(x / 3.0, y / 7.0) for x in range(-9, 9) for y in range(-9, 9)]
        for num in nums:
            self.assertAlmostEqual((num.real ** 2 + num.imag ** 2) ** 0.5, abs(num))

    def test_repr_str(self):
        def test(v, expected, test_fn=self.assertEqual):
            test_fn(repr(v), expected)
            test_fn(str(v), expected)

        test(1 + 6j, "(1+6j)")
        test(1 - 6j, "(1-6j)")

        test(-(1 + 0j), "(-1+-0j)", test_fn=self.assertNotEqual)

        test(Complex(1.0, INF), "(1+infj)")
        test(Complex(1.0, -INF), "(1-infj)")
        test(Complex(INF, 1), "(inf+1j)")
        test(Complex(-INF, INF), "(-inf+infj)")
        test(Complex(NAN, 1), "(nan+1j)")
        test(Complex(1, NAN), "(1+nanj)")
        test(Complex(NAN, NAN), "(nan+nanj)")

        test(Complex(0, INF), "infj")
        test(Complex(0, -INF), "-infj")
        test(Complex(0, NAN), "nanj")

        self.assertEqual(1 - 6j, Complex(repr(1 - 6j)))
        self.assertEqual(1 + 6j, Complex(repr(1 + 6j)))
        self.assertEqual(-6j, Complex(repr(-6j)))
        self.assertEqual(6j, Complex(repr(6j)))

    # @support.requires_IEEE_754
    def test_negative_zero_repr_str(self):
        def test(v, expected, test_fn=self.assertEqual):
            test_fn(repr(v), expected)
            test_fn(str(v), expected)

        test(Complex(0.0, 1.0), "1j")
        test(Complex(-0.0, 1.0), "(-0+1j)")
        test(Complex(0.0, -1.0), "-1j")
        test(Complex(-0.0, -1.0), "(-0-1j)")

        test(Complex(0.0, 0.0), "0j")
        test(Complex(0.0, -0.0), "-0j")
        test(Complex(-0.0, 0.0), "(-0+0j)")
        test(Complex(-0.0, -0.0), "(-0-0j)")

    def test_neg(self):
        self.assertEqual(-(1 + 6j), -1 - 6j)

    # FIXME doesn't work
    # def test_file(self):
    #     a = 3.33 + 4.43j
    #     b = 5.1 + 2.3j

    #     fo = None
    #     try:
    #         fo = open(support.TESTFN, "w")
    #         print(a, b, file=fo)
    #         fo.close()
    #         fo = open(support.TESTFN, "r")
    #         self.assertEqual(fo.read(), ("%s %s\n" % (a, b)))
    #     finally:
    #         if (fo is not None) and (not fo.closed):
    #             fo.close()
    #         support.unlink(support.TESTFN)

    def test_getnewargs(self):
        self.assertEqual((1 + 2j).__getnewargs__(), (1.0, 2.0))
        self.assertEqual((1 - 2j).__getnewargs__(), (1.0, -2.0))
        self.assertEqual((2j).__getnewargs__(), (0.0, 2.0))
        self.assertEqual((-0j).__getnewargs__(), (0.0, -0.0))
        self.assertEqual(Complex(0, INF).__getnewargs__(), (0.0, INF))
        self.assertEqual(Complex(INF, 0).__getnewargs__(), (INF, 0.0))

    # @support.requires_IEEE_754
    def test_plus_minus_0j(self):
        # test that -0j and 0j literals are not identified
        z1, z2 = 0j, -0j
        self.assertEqual(atan2(z1.imag, -1.0), atan2(0.0, -1.0))
        self.assertEqual(atan2(z2.imag, -1.0), atan2(-0.0, -1.0))

    # @support.requires_IEEE_754
    def test_negated_imaginary_literal(self):
        z0 = -0j
        z1 = -7j
        z2 = -1e1000j
        # Note: In versions of Python < 3.2, a negated imaginary literal
        # accidentally ended up with real part 0.0 instead of -0.0, thanks to a
        # modification during CST -> AST translation (see issue #9011).  That's
        # fixed in Python 3.2.
        self.assertFloatsAreIdentical(z0.real, -0.0)
        self.assertFloatsAreIdentical(z0.imag, -0.0)
        self.assertFloatsAreIdentical(z1.real, -0.0)
        self.assertFloatsAreIdentical(z1.imag, -7.0)
        self.assertFloatsAreIdentical(z2.real, -0.0)
        self.assertFloatsAreIdentical(z2.imag, -INF)

    # @support.requires_IEEE_754
    def test_overflow(self):
        self.assertEqual(Complex("1e500"), Complex(INF, 0.0))
        self.assertEqual(Complex("-1e500j"), Complex(0.0, -INF))
        self.assertEqual(Complex("-1e500+1.8e308j"), Complex(-INF, INF))

    # @support.requires_IEEE_754
    def test_repr_roundtrip(self):
        vals = [0.0, 1e-500, 1e-315, 1e-200, 0.0123, 3.1415, 1e50, INF, NAN]
        vals += [-v for v in vals]

        # Complex(repr(z)) should recover z exactly, even for Complex
        # numbers involving an infinity, nan, or negative zero
        for x in vals:
            for y in vals:
                z = Complex(x, y)
                roundtrip = Complex(repr(z))
                self.assertFloatsAreIdentical(z.real, roundtrip.real)
                self.assertFloatsAreIdentical(z.imag, roundtrip.imag)

        # if we predefine some constants, then eval(repr(z)) should
        # also work, except that it might change the sign of zeros
        inf, nan = float("inf"), float("nan")
        infj, nanj = Complex(0.0, inf), Complex(0.0, nan)  # noqa: F841
        for x in vals:
            for y in vals:
                z = Complex(x, y)
                roundtrip = eval(repr(z))
                # adding 0.0 has no effect beside changing -0.0 to 0.0
                self.assertFloatsAreIdentical(0.0 + z.real, 0.0 + roundtrip.real)
                self.assertFloatsAreIdentical(0.0 + z.imag, 0.0 + roundtrip.imag)

    def test_format(self):
        # empty format string is same as str()
        self.assertEqual(format(1 + 3j, ""), str(1 + 3j))
        self.assertEqual(format(1.5 + 3.5j, ""), str(1.5 + 3.5j))
        self.assertEqual(format(3j, ""), str(3j))
        self.assertEqual(format(3.2j, ""), str(3.2j))
        self.assertEqual(format(3 + 0j, ""), str(3 + 0j))
        self.assertEqual(format(3.2 + 0j, ""), str(3.2 + 0j))

        # empty presentation type should still be analogous to str,
        # even when format string is nonempty (issue #5920).
        self.assertEqual(format(3.2 + 0j, "-"), str(3.2 + 0j))
        self.assertEqual(format(3.2 + 0j, "<"), str(3.2 + 0j))
        z = 4 / 7.0 - 100j / 7.0
        self.assertEqual(format(z, ""), str(z))
        self.assertEqual(format(z, "-"), str(z))
        self.assertEqual(format(z, "<"), str(z))
        self.assertEqual(format(z, "10"), str(z))
        z = Complex(0.0, 3.0)
        self.assertEqual(format(z, ""), str(z))
        self.assertEqual(format(z, "-"), str(z))
        self.assertEqual(format(z, "<"), str(z))
        self.assertEqual(format(z, "2"), str(z))
        z = Complex(-0.0, 2.0)
        self.assertEqual(format(z, ""), str(z))
        self.assertEqual(format(z, "-"), str(z))
        self.assertEqual(format(z, "<"), str(z))
        self.assertEqual(format(z, "3"), str(z))

        self.assertEqual(format(1 + 3j, "g"), "1+3j")
        self.assertEqual(format(3j, "g"), "0+3j")
        self.assertEqual(format(1.5 + 3.5j, "g"), "1.5+3.5j")

        self.assertEqual(format(1.5 + 3.5j, "+g"), "+1.5+3.5j")
        self.assertEqual(format(1.5 - 3.5j, "+g"), "+1.5-3.5j")
        self.assertEqual(format(1.5 - 3.5j, "-g"), "1.5-3.5j")
        self.assertEqual(format(1.5 + 3.5j, " g"), " 1.5+3.5j")
        self.assertEqual(format(1.5 - 3.5j, " g"), " 1.5-3.5j")
        self.assertEqual(format(-1.5 + 3.5j, " g"), "-1.5+3.5j")
        self.assertEqual(format(-1.5 - 3.5j, " g"), "-1.5-3.5j")

        self.assertEqual(format(-1.5 - 3.5e-20j, "g"), "-1.5-3.5e-20j")
        self.assertEqual(format(-1.5 - 3.5j, "f"), "-1.500000-3.500000j")
        self.assertEqual(format(-1.5 - 3.5j, "F"), "-1.500000-3.500000j")
        self.assertEqual(format(-1.5 - 3.5j, "e"), "-1.500000e+00-3.500000e+00j")
        self.assertEqual(format(-1.5 - 3.5j, ".2e"), "-1.50e+00-3.50e+00j")
        self.assertEqual(format(-1.5 - 3.5j, ".2E"), "-1.50E+00-3.50E+00j")
        self.assertEqual(format(-1.5e10 - 3.5e5j, ".2G"), "-1.5E+10-3.5E+05j")

        self.assertEqual(format(1.5 + 3j, "<20g"), "1.5+3j              ")
        self.assertEqual(format(1.5 + 3j, "*<20g"), "1.5+3j**************")
        self.assertEqual(format(1.5 + 3j, ">20g"), "              1.5+3j")
        self.assertEqual(format(1.5 + 3j, "^20g"), "       1.5+3j       ")
        self.assertEqual(format(1.5 + 3j, "<20"), "(1.5+3j)            ")
        self.assertEqual(format(1.5 + 3j, ">20"), "            (1.5+3j)")
        self.assertEqual(format(1.5 + 3j, "^20"), "      (1.5+3j)      ")
        self.assertEqual(format(1.123 - 3.123j, "^20.2"), "     (1.1-3.1j)     ")

        self.assertEqual(format(1.5 + 3j, "20.2f"), "          1.50+3.00j")
        self.assertEqual(format(1.5 + 3j, ">20.2f"), "          1.50+3.00j")
        self.assertEqual(format(1.5 + 3j, "<20.2f"), "1.50+3.00j          ")
        self.assertEqual(
            format(1.5e20 + 3j, "<20.2f"), "150000000000000000000.00+3.00j"
        )
        self.assertEqual(
            format(1.5e20 + 3j, ">40.2f"), "          150000000000000000000.00+3.00j"
        )
        self.assertEqual(
            format(1.5e20 + 3j, "^40,.2f"), "  150,000,000,000,000,000,000.00+3.00j  "
        )
        self.assertEqual(
            format(1.5e21 + 3j, "^40,.2f"), " 1,500,000,000,000,000,000,000.00+3.00j "
        )
        self.assertEqual(
            format(1.5e21 + 3000j, ",.2f"), "1,500,000,000,000,000,000,000.00+3,000.00j"
        )

        # Issue 7094: Alternate formatting (specified by #)
        self.assertEqual(format(1 + 1j, ".0e"), "1e+00+1e+00j")
        self.assertEqual(format(1 + 1j, "#.0e"), "1.e+00+1.e+00j")
        self.assertEqual(format(1 + 1j, ".0f"), "1+1j")
        self.assertEqual(format(1 + 1j, "#.0f"), "1.+1.j")
        self.assertEqual(format(1.1 + 1.1j, "g"), "1.1+1.1j")
        self.assertEqual(format(1.1 + 1.1j, "#g"), "1.10000+1.10000j")

        # Alternate doesn't make a difference for these, they format the same with or without it
        self.assertEqual(format(1 + 1j, ".1e"), "1.0e+00+1.0e+00j")
        self.assertEqual(format(1 + 1j, "#.1e"), "1.0e+00+1.0e+00j")
        self.assertEqual(format(1 + 1j, ".1f"), "1.0+1.0j")
        self.assertEqual(format(1 + 1j, "#.1f"), "1.0+1.0j")

        # Misc. other alternate tests
        self.assertEqual(format((-1.5 + 0.5j), "#f"), "-1.500000+0.500000j")
        self.assertEqual(format((-1.5 + 0.5j), "#.0f"), "-2.+0.j")
        self.assertEqual(format((-1.5 + 0.5j), "#e"), "-1.500000e+00+5.000000e-01j")
        self.assertEqual(format((-1.5 + 0.5j), "#.0e"), "-2.e+00+5.e-01j")
        self.assertEqual(format((-1.5 + 0.5j), "#g"), "-1.50000+0.500000j")
        self.assertEqual(format((-1.5 + 0.5j), ".0g"), "-2+0.5j")
        self.assertEqual(format((-1.5 + 0.5j), "#.0g"), "-2.+0.5j")

        # zero padding is invalid
        self.assertRaises(ValueError, (1.5 + 0.5j).__format__, "010f")

        # '=' alignment is invalid
        self.assertRaises(ValueError, (1.5 + 3j).__format__, "=20")

        # integer presentation types are an error
        for t in "bcdoxX":
            self.assertRaises(ValueError, (1.5 + 0.5j).__format__, t)

        # make sure everything works in ''.format()
        self.assertEqual(f"*{3.14159 + 2.71828j:.3f}*", "*3.142+2.718j*")

        # issue 3382
        self.assertEqual(format(Complex(NAN, NAN), "f"), "nan+nanj")
        self.assertEqual(format(Complex(1, NAN), "f"), "1.000000+nanj")
        self.assertEqual(format(Complex(NAN, 1), "f"), "nan+1.000000j")
        self.assertEqual(format(Complex(NAN, -1), "f"), "nan-1.000000j")
        self.assertEqual(format(Complex(NAN, NAN), "F"), "NAN+NANj")
        self.assertEqual(format(Complex(1, NAN), "F"), "1.000000+NANj")
        self.assertEqual(format(Complex(NAN, 1), "F"), "NAN+1.000000j")
        self.assertEqual(format(Complex(NAN, -1), "F"), "NAN-1.000000j")
        self.assertEqual(format(Complex(INF, INF), "f"), "inf+infj")
        self.assertEqual(format(Complex(1, INF), "f"), "1.000000+infj")
        self.assertEqual(format(Complex(INF, 1), "f"), "inf+1.000000j")
        self.assertEqual(format(Complex(INF, -1), "f"), "inf-1.000000j")
        self.assertEqual(format(Complex(INF, INF), "F"), "INF+INFj")
        self.assertEqual(format(Complex(1, INF), "F"), "1.000000+INFj")
        self.assertEqual(format(Complex(INF, 1), "F"), "INF+1.000000j")
        self.assertEqual(format(Complex(INF, -1), "F"), "INF-1.000000j")



# stdlib
from typing import Any
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft relative
from ...proto.core.common.common_object_pb2 import ObjectWithID as ObjectWithID_PB
from ...util import validate_type
from ..common.serde.deserialize import _deserialize
from ..common.serde.serializable import Serializable
from ..common.serde.serializable import bind_protobuf
from ..common.serde.serialize import _serialize as serialize
from .uid import UID


@bind_protobuf
class ObjectWithID(Serializable):
    """This object is the superclass for nearly all Syft objects. Subclassing
    from this object will cause an object to be initialized with a unique id
    using the process specified in the UID class.

    .. note::
        At the time of writing, the only class in Syft which doesn't have an ID
        of some kind is the Client class because it's job is to point to another
        object (which has an ID).

    .. note::
        Be aware of performance choices in this class because it is used so
        heavily across the entire codebase. Assume every method is going to
        be called thousands of times during the working day of an average
        data scientist using syft (and millions of times in the context of a
        machine learning job).

    """

    def __init__(self, id: Optional[UID] = None):
        """This initializer only exists to set the id attribute, which is the
        primary purpose of this class. It also sets the 'as_wrapper' flag
        for the 'Serializable' superclass.

        Args:
            id: an override which can be used to set an ID for this object

        """

        if id is None:
            id = UID()

        self._id: UID = id

        # while this class is never used as a simple wrapper,
        # it's possible that sub-classes of this class will be.
        super().__init__()

    @property
    def id(self) -> UID:
        """We reveal ObjectWithID.id as a property to discourage users and
        developers of Syft from modifying .id attributes after an object
        has been initialized.

        Returns:
            returns the unique id of the object
        """
        return self._id

    def __eq__(self, other: Any) -> bool:
        """Checks to see if two ObjectWithIDs are actually the same object.

        This checks to see whether this ObjectWithIDs is equal to another by
        comparing whether they have the same .id objects. These objects
        come with their own __eq__ function which we assume to be correct.

        Args:
            other: this is the other ObjectWithIDs to be compared with

        Returns:
            True/False based on whether the objects are the same
        """

        try:
            return self.id == other.id
        except Exception:
            return False

    def __repr__(self) -> str:
        """
        Return a human-readable representation of the ObjectWithID with brackets
        so that it can be easily spotted when nested inside of the human-
        readable representations of other objects.

        Returns:
            a human-readable version of the ObjectWithID

        """

        no_dash = str(self.id.value).replace("-", "")
        return f"<{type(self).__name__}: {no_dash}>"

    def repr_short(self) -> str:
        """
        Return a SHORT human-readable version of the ID which
        makes it print nicer when embedded (often alongside other
        UID objects) within other object __repr__ methods.

        Returns:
            a SHORT human-readable version of SpecificLocation
        """

        return f"<{type(self).__name__}:{self.id.repr_short()}>"

    def _object2proto(self) -> ObjectWithID_PB:
        """
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        Returns:
            a protobuf object that is the serialization of self.

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return ObjectWithID_PB(id=serialize(self.id))

    @staticmethod
    def _proto2object(proto: ObjectWithID_PB) -> "ObjectWithID":
        """Creates a ObjectWithID from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        Args:
            proto: a protobuf object that we wish to convert to instance of this class

        Returns:
            an instance of ObjectWithID

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """
        _id = validate_type(_object=_deserialize(proto.id), _type=UID, optional=True)
        return ObjectWithID(id=_id)

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        Returns:
            the type of protobuf object which corresponds to this class.

        """

        return ObjectWithID_PB



# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft relative
from ..... import lib
from ..... import serialize
from .....logger import traceback_and_raise
from .....proto.core.node.common.action.run_function_or_constructor_pb2 import (
    RunFunctionOrConstructorAction as RunFunctionOrConstructorAction_PB,
)
from .....util import inherit_tags
from ....common.serde.deserialize import _deserialize
from ....common.serde.serializable import bind_protobuf
from ....common.uid import UID
from ....io.address import Address
from ....pointer.pointer import Pointer
from ....store.storeable_object import StorableObject
from ...abstract.node import AbstractNode
from ..util import listify
from .common import ImmediateActionWithoutReply


@bind_protobuf
class RunFunctionOrConstructorAction(ImmediateActionWithoutReply):
    """
    When executing a RunFunctionOrConstructorAction, a :class:`Node` will run
    a function defined by the action's path attribute and keep the returned value
    in its store.

    Attributes:
         path: the dotted path to the function to call
         args: args to pass to the function. They should be pointers to objects
            located on the :class:`Node` that will execute the action.
         kwargs: kwargs to pass to the function. They should be pointers to objects
            located on the :class:`Node` that will execute the action.
    """

    def __init__(
        self,
        path: str,
        args: Union[Tuple[Any, ...], List[Any]],
        kwargs: Dict[Any, Any],
        id_at_location: UID,
        address: Address,
        msg_id: Optional[UID] = None,
        is_static: Optional[bool] = False,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.path = path
        self.args = listify(args)  # args need to be editable for plans
        self.kwargs = kwargs
        self.id_at_location = id_at_location
        self.is_static = is_static

    @staticmethod
    def intersect_keys(
        left: Union[Dict[VerifyKey, UID], None], right: Dict[VerifyKey, UID]
    ) -> Dict[VerifyKey, UID]:
        # FIXME duplicated in run_class_method_action.py
        # get the intersection of the dict keys, the value is the request_id
        # if the request_id is different for some reason we still want to keep it,
        # so only intersect the keys and then copy those over from the main dict
        # into a new one
        if left is None:
            return right
        intersection = left.keys() & right.keys()
        # left and right have the same keys
        return {k: left[k] for k in intersection}

    def execute_action(self, node: AbstractNode, verify_key: VerifyKey) -> None:
        method = node.lib_ast(self.path)
        result_read_permissions: Union[None, Dict[VerifyKey, UID]] = None

        resolved_args = list()
        tag_args = []
        for arg in self.args:
            if not isinstance(arg, Pointer):
                traceback_and_raise(
                    ValueError(
                        f"args attribute of RunFunctionOrConstructorAction should only contain Pointers. "
                        f"Got {arg} of type {type(arg)}"
                    )
                )

            r_arg = node.store[arg.id_at_location]
            result_read_permissions = self.intersect_keys(
                result_read_permissions, r_arg.read_permissions
            )
            resolved_args.append(r_arg.data)
            tag_args.append(r_arg)

        resolved_kwargs = {}
        tag_kwargs = {}
        for arg_name, arg in self.kwargs.items():
            if not isinstance(arg, Pointer):
                traceback_and_raise(
                    ValueError(
                        f"kwargs attribute of RunFunctionOrConstructorAction should only contain Pointers. "
                        f"Got {arg} of type {type(arg)}"
                    )
                )

            r_arg = node.store[arg.id_at_location]
            result_read_permissions = self.intersect_keys(
                result_read_permissions, r_arg.read_permissions
            )
            resolved_kwargs[arg_name] = r_arg.data
            tag_kwargs[arg_name] = r_arg

        # upcast our args in case the method only accepts the original types
        (
            upcasted_args,
            upcasted_kwargs,
        ) = lib.python.util.upcast_args_and_kwargs(resolved_args, resolved_kwargs)

        # execute the method with the newly upcasted args and kwargs
        result = method(*upcasted_args, **upcasted_kwargs)

        # to avoid circular imports
        if lib.python.primitive_factory.isprimitive(value=result):
            # Wrap in a SyPrimitive
            result = lib.python.primitive_factory.PrimitiveFactory.generate_primitive(
                value=result, id=self.id_at_location
            )
        else:
            if hasattr(result, "id"):
                result._id = self.id_at_location

        # If we have no permission (None or {}) we add some default permissions based on a permission list
        if result_read_permissions is None:
            result_read_permissions = {}

        if not isinstance(result, StorableObject):
            result = StorableObject(
                id=self.id_at_location,
                data=result,
                read_permissions=result_read_permissions,
            )

        inherit_tags(
            attr_path_and_name=self.path,
            result=result,
            self_obj=None,
            args=tag_args,
            kwargs=tag_kwargs,
        )

        node.store[self.id_at_location] = result

    def __repr__(self) -> str:
        method_name = self.path.split(".")[-1]
        arg_names = ",".join([a.class_name for a in self.args])
        kwargs_names = ",".join([f"{k}={v.class_name}" for k, v in self.kwargs.items()])
        return f"RunClassMethodAction {method_name}({arg_names}, {kwargs_names})"

    def _object2proto(self) -> RunFunctionOrConstructorAction_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: RunFunctionOrConstructorAction_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return RunFunctionOrConstructorAction_PB(
            path=self.path,
            args=[serialize(x) for x in self.args],
            kwargs={k: serialize(v) for k, v in self.kwargs.items()},
            id_at_location=serialize(self.id_at_location),
            address=serialize(self.address),
            msg_id=serialize(self.id),
        )

    @staticmethod
    def _proto2object(
        proto: RunFunctionOrConstructorAction_PB,
    ) -> "RunFunctionOrConstructorAction":
        """Creates a ObjectWithID from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of RunFunctionOrConstructorAction
        :rtype: RunFunctionOrConstructorAction

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return RunFunctionOrConstructorAction(
            path=proto.path,
            args=tuple(_deserialize(blob=x) for x in proto.args),
            kwargs={k: _deserialize(blob=v) for k, v in proto.kwargs.items()},
            id_at_location=_deserialize(blob=proto.id_at_location),
            address=_deserialize(blob=proto.address),
            msg_id=_deserialize(blob=proto.msg_id),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return RunFunctionOrConstructorAction_PB

    def remap_input(self, current_input: Any, new_input: Any) -> None:
        """Redefines some of the arguments of the function"""
        for i, arg in enumerate(self.args):
            if arg.id_at_location == current_input.id_at_location:
                self.args[i] = new_input

        for k, v in self.kwargs.items():
            if v.id_at_location == current_input.id_at_location:
                self.kwargs[k] = new_input



# stdlib
from typing import Callable
from typing import Union

# third party
from nacl.bindings.crypto_sign import crypto_sign_keypair
from nacl.signing import VerifyKey

# syft absolute
from syft.core.common.message import SignedEventualSyftMessageWithoutReply
from syft.core.common.message import SignedImmediateSyftMessageWithReply
from syft.core.common.message import SignedImmediateSyftMessageWithoutReply
from syft.core.common.uid import UID
from syft.core.io.address import Address
from syft.core.io.location.specific import SpecificLocation
from syft.core.node.common.node import Node


class MockNode(Node):
    """Mock Node object for testing purposes."""

    def recv_immediate_msg_without_reply(
        self, msg: SignedImmediateSyftMessageWithoutReply
    ) -> None:
        return None

    def recv_eventual_msg_without_reply(
        self, msg: SignedImmediateSyftMessageWithReply
    ) -> None:
        return None

    def recv_immediate_msg_with_reply(
        self, msg: SignedImmediateSyftMessageWithReply
    ) -> SignedImmediateSyftMessageWithoutReply:
        return construct_dummy_message(SignedImmediateSyftMessageWithoutReply)


def _construct_address() -> Address:
    """Helper method to construct an Address"""
    return Address(
        network=SpecificLocation(id=UID()),
        domain=SpecificLocation(id=UID()),
        device=SpecificLocation(id=UID()),
        vm=SpecificLocation(id=UID()),
    )


def construct_dummy_message(
    msg_class: Callable,
) -> Union[
    SignedEventualSyftMessageWithoutReply,
    SignedImmediateSyftMessageWithReply,
    SignedImmediateSyftMessageWithoutReply,
]:
    """
    Helper method to construct a dummy SignedMessage of the following:
    - SignedEventualSyftMessageWithoutReply
    - SignedImmediateSyftMessageWithReply
    - SignedImmediateSyftMessageWithoutReply
    """
    address = _construct_address()
    key = VerifyKey(crypto_sign_keypair()[0])

    return msg_class(
        address=address,
        obj_type="signed_message",
        signature=b"my_signature",
        verify_key=key,
        message=b"hello_world",
    )



# third party
import zksk as zk  # noqa: 401

# syft relative
from ...core.common.serde.deserialize import _deserialize as deserialize
from ...core.common.serde.serialize import _serialize as serialize
from ...generate_wrapper import GenerateWrapper
from ...lib.python import Tuple
from ...proto.lib.zksk.nizk_pb2 import NIZK as NIZK_PB


def object2proto(obj: zk.base.NIZK) -> NIZK_PB:
    return NIZK_PB(
        challenge=serialize(obj.challenge, to_proto=True),
        responses=serialize(Tuple(obj.responses), to_proto=True),
        stmt_hash=obj.stmt_hash,
    )


def proto2object(proto: NIZK_PB) -> zk.expr.Secret:
    return zk.base.NIZK(
        challenge=deserialize(proto.challenge, from_proto=True),
        responses=deserialize(proto.responses, from_proto=True).upcast(),
        stmt_hash=proto.stmt_hash,
    )


GenerateWrapper(
    wrapped_type=zk.base.NIZK,
    import_path="zksk.base.NIZK",
    protobuf_scheme=NIZK_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# syft relative
from .dataset import Dataset
from .store_disk import DiskObjectStore
from .store_interface import ObjectStore
from .store_memory import MemoryStore

__all__ = ["DiskObjectStore", "ObjectStore", "MemoryStore", "Dataset"]



# stdlib
from typing import Any
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft relative
from ... import deserialize
from ... import serialize
from ...core.common import UID
from ...core.common.serde.serializable import bind_protobuf
from ...proto.lib.python.slice_pb2 import Slice as Slice_PB
from .primitive_factory import PrimitiveFactory
from .primitive_interface import PyPrimitive
from .types import SyPrimitiveRet


@bind_protobuf
class Slice(PyPrimitive):
    def __init__(
        self,
        start: Any = None,
        stop: Optional[Any] = None,
        step: Optional[Any] = None,
        id: Optional[UID] = None,
    ):
        # first, second, third
        if stop is None and step is None:
            # slice treats 1 arg as stop not start
            stop = start
            start = None

        self.value = slice(start, stop, step)
        self._id: UID = id if id else UID()

    @property
    def id(self) -> UID:
        """We reveal PyPrimitive.id as a property to discourage users and
        developers of Syft from modifying .id attributes after an object
        has been initialized.

        :return: returns the unique id of the object
        :rtype: UID
        """
        return self._id

    def __eq__(self, other: Any) -> SyPrimitiveRet:
        res = self.value.__eq__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __ge__(self, other: Any) -> SyPrimitiveRet:
        res = self.value.__ge__(other)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def __gt__(self, other: Any) -> SyPrimitiveRet:
        res = self.value.__gt__(other)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def __le__(self, other: Any) -> SyPrimitiveRet:
        res = self.value.__le__(other)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def __lt__(self, other: Any) -> SyPrimitiveRet:
        res = self.value.__lt__(other)  # type: ignore
        return PrimitiveFactory.generate_primitive(value=res)

    def __ne__(self, other: Any) -> SyPrimitiveRet:
        res = self.value.__ne__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __repr__(self) -> str:
        return self.value.__repr__()

    def __str__(self) -> str:
        return self.value.__str__()

    def indices(self, index: int) -> tuple:
        res = self.value.indices(index)
        return PrimitiveFactory.generate_primitive(value=res)

    @property
    def start(self) -> Optional[int]:
        return self.value.start

    @property
    def step(self) -> Optional[int]:
        return self.value.step

    @property
    def stop(self) -> Optional[int]:
        return self.value.stop

    def upcast(self) -> slice:
        return self.value

    def _object2proto(self) -> Slice_PB:
        slice_pb = Slice_PB()
        if self.start:
            slice_pb.start = self.start
            slice_pb.has_start = True

        if self.stop:
            slice_pb.stop = self.stop
            slice_pb.has_stop = True

        if self.step:
            slice_pb.step = self.step
            slice_pb.has_step = True

        slice_pb.id.CopyFrom(serialize(obj=self._id))

        return slice_pb

    @staticmethod
    def _proto2object(proto: Slice_PB) -> "Slice":
        id_: UID = deserialize(blob=proto.id)
        start = None
        stop = None
        step = None
        if proto.has_start:
            start = proto.start

        if proto.has_stop:
            stop = proto.stop

        if proto.has_step:
            step = proto.step

        return Slice(
            start=start,
            stop=stop,
            step=step,
            id=id_,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return Slice_PB



# stdlib
import gc

# third party
import pytest
import torch as th

# syft absolute
import syft as sy


@pytest.mark.slow
def test_parameter_vm_remote_operation(
    node: sy.VirtualMachine, client: sy.VirtualMachineClient
) -> None:
    x = th.nn.Parameter(th.randn(3, 3))

    xp = x.send(client, pointable=False)

    y = xp + xp

    assert len(node.store._objects) == 2

    y.get()

    assert len(node.store._objects) == 1

    del xp

    gc.collect()

    assert len(node.store._objects) == 0


def test_get_copy(node: sy.VirtualMachine, client: sy.VirtualMachineClient) -> None:
    x = th.nn.Parameter(th.randn(3, 3))

    xp = x.send(client, pointable=False)

    y = xp + xp

    assert len(node.store._objects) == 2

    y.get_copy()

    # no deletion of the object
    assert len(node.store._objects) == 2

    del xp
    gc.collect()

    assert len(node.store._objects) == 1


def test_parameter_serde() -> None:
    param = th.nn.parameter.Parameter(th.tensor([1.0, 2, 3]), requires_grad=True)
    # Setting grad manually to check it is passed through serialization
    param.grad = th.randn_like(param)

    blob = sy.serialize(param)

    param2 = sy.deserialize(blob=blob)

    assert (param == param2).all()
    assert (param2.grad == param2.grad).all()
    assert param2.requires_grad == param2.requires_grad


def test_linear_grad_serde() -> None:
    # Parameter is created inside Linear module
    linear = th.nn.Linear(5, 1)
    param = linear.weight

    # we dont have auto generated IDs anymore
    # assert hasattr(param, "id")

    # TODO: fix backward is failing
    # Induce grads on linear weights
    # out = linear(th.randn(5, 5))
    # out.backward()
    # assert param.grad is not None

    blob = sy.serialize(param)

    param2 = sy.deserialize(blob=blob)

    assert (param == param2).all()
    # assert param == param2
    # assert (param2.grad == param2.grad).all()
    assert param2.requires_grad == param2.requires_grad



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/io/route.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.io import (
    connection_pb2 as proto_dot_core_dot_io_dot_connection__pb2,
)
from syft.proto.core.io import location_pb2 as proto_dot_core_dot_io_dot_location__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/core/io/route.proto",
    package="syft.core.io",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x19proto/core/io/route.proto\x12\x0csyft.core.io\x1a\x1cproto/core/io/location.proto\x1a\x1eproto/core/io/connection.proto"{\n\tSoloRoute\x12\x33\n\x0b\x64\x65stination\x18\x01 \x01(\x0b\x32\x1e.syft.core.io.SpecificLocation\x12\x39\n\nconnection\x18\x02 \x01(\x0b\x32%.syft.core.io.VirtualClientConnectionb\x06proto3',
    dependencies=[
        proto_dot_core_dot_io_dot_location__pb2.DESCRIPTOR,
        proto_dot_core_dot_io_dot_connection__pb2.DESCRIPTOR,
    ],
)


_SOLOROUTE = _descriptor.Descriptor(
    name="SoloRoute",
    full_name="syft.core.io.SoloRoute",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="destination",
            full_name="syft.core.io.SoloRoute.destination",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="connection",
            full_name="syft.core.io.SoloRoute.connection",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=105,
    serialized_end=228,
)

_SOLOROUTE.fields_by_name[
    "destination"
].message_type = proto_dot_core_dot_io_dot_location__pb2._SPECIFICLOCATION
_SOLOROUTE.fields_by_name[
    "connection"
].message_type = proto_dot_core_dot_io_dot_connection__pb2._VIRTUALCLIENTCONNECTION
DESCRIPTOR.message_types_by_name["SoloRoute"] = _SOLOROUTE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SoloRoute = _reflection.GeneratedProtocolMessageType(
    "SoloRoute",
    (_message.Message,),
    {
        "DESCRIPTOR": _SOLOROUTE,
        "__module__": "proto.core.io.route_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.io.SoloRoute)
    },
)
_sym_db.RegisterMessage(SoloRoute)


# @@protoc_insertion_point(module_scope)



# stdlib
import multiprocessing as mp
from multiprocessing import Process
import time

# third party
import torch as th

mp.set_start_method("spawn", force=True)

# Make sure to run the local network.py server first:
#
# $ syft-network
#


def do() -> None:
    # syft absolute
    import syft as sy

    _ = sy.logger.add(
        sink="syft_do.log",
        level="TRACE",
    )

    duet = sy.launch_duet(loopback=True, network_url="http://localhost:5000")
    duet.requests.add_handler(action="accept")
    t = th.randn(4000000)
    print("DO: Tensor sum:", t.sum())
    start = time.time()
    tp = t.send(duet, pointable=True)
    end = time.time()
    print("DO: Pointer: ", tp, "serialized in", end - start)
    print("DO: Store: ", duet.store)
    sy.core.common.event_loop.loop.run_forever()
    print("DO: DONE")


def ds() -> None:
    # syft absolute
    import syft as sy

    _ = sy.logger.add(
        sink="syft_ds.log",
        level="TRACE",
    )

    duet = sy.join_duet(loopback=True, network_url="http://localhost:5000")
    time.sleep(1)
    print("DS: Store: ", duet.store)
    start = time.time()
    t = duet.store[0].get(request_block=True, delete_obj=False)
    end = time.time()
    print("DS: Received in:", end - start)
    print("DS: Shape: ", t.shape)
    print("DS: Tensor sum:", t.sum())


if __name__ == "__main__":
    p1 = Process(target=do)
    p1.start()
    p2 = Process(target=ds)
    p2.start()
    p2.join()
    p1.terminate()



# future
from __future__ import print_function

# stdlib
import argparse
from math import log10

# third party
from model import Net
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader

from data import get_test_set  # isort:skip
from data import get_training_set  # isort:skip

# Training settings
parser = argparse.ArgumentParser(description="PyTorch Super Res Example")
parser.add_argument(
    "--upscale-factor",
    type=int,
    required=True,
    help="super resolution upscale factor[REQUIRED]"
)
parser.add_argument(
    "--batch-size",
    type=int,
    default=64,
    help="training batch size; default: 64"
)
parser.add_argument(
    "--test-batch-size",
    type=int,
    default=10,
    help="testing batch size; default: 10"
)
parser.add_argument(
    "--epochs",
    type=int,
    default=2,
    help="number of epochs to train; default: 2"
)
parser.add_argument(
    "--lr",
    type=float,
    default=0.01,
    help="Learning Rate; default: 0.01"
)
parser.add_argument(
    "--cuda",
    action="store_true",
    help="use cuda?")
parser.add_argument(
    "--threads",
    type=int,
    default=4,
    help="number of threads for data loader to use; default: 4"
)
parser.add_argument(
    "--seed",
    type=int,
    default=123,
    help="random seed to use; default:123"
)
opt = parser.parse_args()

print(opt)

if opt.cuda and not torch.cuda.is_available():
    raise Exception("No GPU found, please run without --cuda")

torch.manual_seed(opt.seed)

device = torch.device("cuda" if opt.cuda else "cpu")

print("===> Loading datasets")
train_set = get_training_set(opt.upscale_factor)
test_set = get_test_set(opt.upscale_factor)
training_data_loader = DataLoader(
    dataset=train_set, num_workers=opt.threads, batch_size=opt.batch_size, shuffle=True
)
testing_data_loader = DataLoader(
    dataset=test_set,
    num_workers=opt.threads,
    batch_size=opt.test_batch_size,
    shuffle=False,
)

print("===> Building model")
model = Net(upscale_factor=opt.upscale_factor).to(device)
criterion = nn.MSELoss()

optimizer = optim.Adam(model.parameters(), lr=opt.lr)


def train(epoch):
    epoch_loss = 0
    for iteration, batch in enumerate(training_data_loader, 1):
        input, target = batch[0].to(device), batch[1].to(device)

        optimizer.zero_grad()
        loss = criterion(model(input), target)
        epoch_loss += loss.item()
        loss.backward()
        optimizer.step()

        print(
            "===> Epoch[{}]({}/{}): Loss: {:.4f}".format(
                epoch, iteration, len(training_data_loader), loss.item()
            )
        )

    print(
        "===> Epoch {} Complete: Avg. Loss: {:.4f}".format(
            epoch, epoch_loss / len(training_data_loader)
        )
    )


def test():
    avg_psnr = 0
    with torch.no_grad():
        for batch in testing_data_loader:
            input, target = batch[0].to(device), batch[1].to(device)

            prediction = model(input)
            mse = criterion(prediction, target)
            psnr = 10 * log10(1 / mse.item())
            avg_psnr += psnr
    print(f"===> Avg. PSNR: {avg_psnr / len(testing_data_loader):.4f} dB")


def checkpoint(epoch):
    model_out_path = f"model_epoch_{epoch}.pth"
    torch.save(model, model_out_path)
    print(f"Checkpoint saved to {model_out_path}")


for epoch in range(1, opt.epochs + 1):
    train(epoch)
    test()
    checkpoint(epoch)



# stdlib
from multiprocessing import Pipe
from multiprocessing import Process
import os
import sys
import traceback
from typing import Any
from typing import Optional


class SyftTestProcess(Process):
    def __init__(self, *args: Any, **kwargs: Any):
        Process.__init__(self, *args, **kwargs)
        self._pconn, self._cconn = Pipe()
        self._exception = None

    def run(self) -> None:
        try:
            sys.stdout = open(os.devnull, "w")
            Process.run(self)
            self._cconn.send(None)
        except Exception as e:
            tb = traceback.format_exc()
            self._cconn.send((e, tb))

    @property
    def exception(self) -> Optional[tuple]:
        if self._pconn.poll():
            self._exception = self._pconn.recv()
        return self._exception



# third party
import pytest

# syft absolute
import syft as sy
from syft.grid.duet.ui import LOGO_URL


@pytest.mark.vendor(lib="PIL")
def test_send_and_get(root_client: sy.VirtualMachineClient) -> None:
    # third party
    import PIL

    sy.load("PIL")

    im = PIL.Image.open(LOGO_URL)
    remote_im = im.send(root_client)
    received_im = remote_im.get()

    assert PIL.ImageChops.difference(im, received_im).getbbox() is None


@pytest.mark.vendor(lib="PIL")
def test_remote_create(root_client: sy.VirtualMachineClient) -> None:
    # third party
    import PIL
    import numpy as np
    import torch

    sy.load("PIL")

    remote_torchvision = root_client.torchvision

    im = PIL.Image.open(LOGO_URL)
    im_array = np.array(im)
    im_tensor = torch.Tensor(im_array).permute(2, 0, 1)
    remote_tensor = im_tensor.send(root_client)
    remote_im = remote_torchvision.transforms.functional.to_pil_image(remote_tensor)
    received_im = remote_im.get()

    assert PIL.ImageChops.difference(im, received_im).getbbox() is None



# third party
import torch as th

# syft absolute
import syft as sy
from syft.lib.python.list import List
from syft.proto.lib.python.list_pb2 import List as List_PB


def test_list_serde() -> None:
    t1 = th.tensor([1, 2])
    t2 = th.tensor([1, 3])

    syft_list = List([t1, t2])

    serialized = syft_list._object2proto()

    assert isinstance(serialized, List_PB)

    deserialized = List._proto2object(proto=serialized)

    assert isinstance(deserialized, List)
    assert deserialized.id == syft_list.id
    for deserialized_el, original_el in zip(deserialized, syft_list):
        assert (deserialized_el == original_el).all()


def test_list_send(client: sy.VirtualMachineClient) -> None:
    t1 = th.tensor([1, 2])
    t2 = th.tensor([1, 3])

    syft_list = List([t1, t2])
    ptr = syft_list.send(client)
    # Check pointer type
    assert ptr.__class__.__name__ == "ListPointer"

    # Check that we can get back the object
    res = ptr.get()
    for res_el, original_el in zip(res, syft_list):
        assert (res_el == original_el).all()



# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import SigningKey
from pandas import DataFrame

# syft relative
from ...core.common.message import EventualSyftMessageWithoutReply
from ...core.common.message import ImmediateSyftMessageWithReply
from ...core.common.message import ImmediateSyftMessageWithoutReply
from ...core.common.message import SignedImmediateSyftMessageWithReply
from ...core.common.message import SignedImmediateSyftMessageWithoutReply
from ...core.common.message import SyftMessage
from ...core.common.serde.serialize import _serialize as serialize  # noqa: F401
from ...core.io.address import Address
from ...core.io.connection import ClientConnection
from ...core.io.location.specific import SpecificLocation
from ...core.io.route import SoloRoute
from ...core.node.common.client import Client
from ...core.node.device.client import DeviceClient
from ...core.node.domain.client import DomainClient
from ...core.node.network.client import NetworkClient
from ...core.node.vm.client import VirtualMachineClient
from ...core.pointer.pointer import Pointer
from ..messages.network_search_message import NetworkSearchMessage
from ..messages.setup_messages import CreateInitialSetUpMessage
from ..messages.setup_messages import GetSetUpMessage
from ..messages.transfer_messages import LoadObjectMessage
from .request_api.association_api import AssociationRequestAPI
from .request_api.group_api import GroupRequestAPI
from .request_api.role_api import RoleRequestAPI
from .request_api.user_api import UserRequestAPI
from .request_api.worker_api import WorkerRequestAPI


def connect(
    url: str,
    conn_type: Type[ClientConnection],
    credentials: Dict = {},
    user_key: Optional[SigningKey] = None,
) -> Any:
    class GridClient(DomainClient):
        def __init__(
            self,
            credentials: Dict,
            url: str,
            conn_type: Type[ClientConnection],
            client_type: Type[Client],
        ) -> None:

            # Use Server metadata
            # to build client route
            self.conn = conn_type(url=url)  # type: ignore
            self.client_type = client_type

            if credentials:
                metadata, _user_key = self.conn.login(credentials=credentials)  # type: ignore
                _user_key = SigningKey(_user_key.encode("utf-8"), encoder=HexEncoder)
            else:
                metadata = self.conn._get_metadata()  # type: ignore
                if not user_key:
                    _user_key = SigningKey.generate()
                else:
                    _user_key = user_key

            (
                spec_location,
                name,
                client_id,
            ) = self.client_type.deserialize_client_metadata_from_node(
                metadata=metadata
            )

            # Create a new Solo Route using the selected connection type
            route = SoloRoute(destination=spec_location, connection=self.conn)

            location_args = self.__route_client_location(
                client_type=self.client_type, location=spec_location
            )

            self.proxy_address: Optional[Address] = None

            # Create a new client using the selected client type
            super().__init__(
                network=location_args[NetworkClient],
                domain=location_args[DomainClient],
                device=location_args[DeviceClient],
                vm=location_args[VirtualMachineClient],
                name=name,
                routes=[route],
                signing_key=_user_key,
            )

            self.groups = GroupRequestAPI(send=self.__perform_grid_request)
            self.users = UserRequestAPI(send=self.__perform_grid_request)
            self.roles = RoleRequestAPI(send=self.__perform_grid_request)
            self.workers = WorkerRequestAPI(
                send=self.__perform_grid_request, domain_client=self
            )
            self.association_requests = AssociationRequestAPI(
                send=self.__perform_grid_request
            )

        def load(
            self, obj_ptr: Type[Pointer], address: Address, searchable: bool = False
        ) -> None:
            content = {
                "address": serialize(address).SerializeToString().decode("ISO-8859-1"),  # type: ignore
                "uid": str(obj_ptr.id_at_location.value),
                "searchable": searchable,
            }
            self.__perform_grid_request(grid_msg=LoadObjectMessage, content=content)

        def initial_setup(self, **kwargs: Any) -> Any:
            return self.__perform_grid_request(
                grid_msg=CreateInitialSetUpMessage, content=kwargs
            )

        def get_setup(self, **kwargs: Any) -> Any:
            return self.__perform_grid_request(grid_msg=GetSetUpMessage, content=kwargs)

        def search(self, query: List, pandas: bool = False) -> Any:
            response = self.__perform_grid_request(
                grid_msg=NetworkSearchMessage, content={"query": query}
            )
            if pandas:
                response = DataFrame(response)

            return response

        def send_immediate_msg_with_reply(
            self,
            msg: Union[
                SignedImmediateSyftMessageWithReply, ImmediateSyftMessageWithReply
            ],
            route_index: int = 0,
        ) -> SyftMessage:
            return super(GridClient, self).send_immediate_msg_with_reply(
                msg=msg, route_index=route_index
            )

        def send_immediate_msg_without_reply(
            self,
            msg: Union[
                SignedImmediateSyftMessageWithoutReply, ImmediateSyftMessageWithoutReply
            ],
            route_index: int = 0,
        ) -> None:
            super(GridClient, self).send_immediate_msg_without_reply(
                msg=msg, route_index=route_index
            )

        def send_eventual_msg_without_reply(
            self,
            msg: EventualSyftMessageWithoutReply,
            route_index: int = 0,
        ) -> None:
            return super(GridClient, self).send_eventual_msg_without_reply(
                msg=msg, route_index=route_index
            )

        def __route_client_location(
            self, client_type: Any, location: SpecificLocation
        ) -> Dict[Any, Any]:
            locations: Dict[Any, Optional[SpecificLocation]] = {
                NetworkClient: None,
                DomainClient: None,
                DeviceClient: None,
                VirtualMachineClient: None,
            }
            locations[client_type] = location
            return locations

        def __perform_grid_request(
            self, grid_msg: Any, content: Optional[Dict[Any, Any]] = None
        ) -> Dict[Any, Any]:
            if content is None:
                content = {}
            signed_msg = self.__build_msg(grid_msg=grid_msg, content=content)
            response = self.send_immediate_msg_with_reply(msg=signed_msg)
            return self.__process_response(response=response)

        def __build_msg(self, grid_msg: Any, content: Dict[Any, Any]) -> Any:
            args = {
                "address": self.address,
                "content": content,
                "reply_to": self.address,
            }
            return grid_msg(**args).sign(signing_key=self.signing_key)

        def __process_response(self, response: SyftMessage) -> Dict[Any, Any]:
            if response.status_code == 200:  # type: ignore
                return response.content  # type: ignore
            else:
                raise Exception(response.content["error"])  # type: ignore

    return GridClient(
        credentials=credentials, url=url, conn_type=conn_type, client_type=DomainClient
    )



# third party
import torch as th

# syft absolute
import syft as sy
from syft import serialize
from syft.core.common import UID
from syft.core.store.dataset import Dataset
from syft.core.store.storeable_object import StorableObject


def test_create_dataset_with_store_obj() -> None:
    id = UID()
    data = UID()
    description = "This is a dummy id"
    tags = ["dummy", "test"]
    obj1 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([1, 2, 3, 4])
    description = "This is a dummy tensor"
    tags = ["dummy", "test"]
    obj2 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = [obj1, obj2]
    description = "This is a dummy tensor"
    tags = ["dummy", "dataset"]
    Dataset(id=id, data=data, description=description, tags=tags)


def test_dataset_search_id() -> None:
    id = UID()
    data = UID()
    description = "This is a dummy id"
    tags = ["dummy", "test"]
    obj1 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([1, 2, 3, 4])
    description = "This is a dummy tensor n1"
    tags = ["dummy", "test"]
    obj2 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([10, 20, 30, 40])
    description = "This is a dummy tensor n2"
    tags = ["dummy", "test"]
    obj3 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = [obj1, obj2, obj3]
    description = "This is a dataset"
    tags = ["dummy", "dataset"]
    dataset_obj = Dataset(id=id, data=data, description=description, tags=tags)

    assert dataset_obj.__contains__(_id=obj1.id)


def test_dataset_search_id_fail() -> None:
    id = UID()
    data = UID()
    description = "This is a dummy id"
    tags = ["dummy", "test"]
    obj1 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([1, 2, 3, 4])
    description = "This is a dummy tensor n1"
    tags = ["dummy", "test"]
    obj2 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([10, 20, 30, 40])
    description = "This is a dummy tensor n2"
    tags = ["dummy", "test"]
    obj3 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = [obj1, obj2, obj3]
    description = "This is a dataset"
    tags = ["dummy", "dataset"]
    dataset_obj = Dataset(id=id, data=data, description=description, tags=tags)

    assert not dataset_obj.__contains__(_id=UID())


def test_dataset_get_element() -> None:
    id = UID()
    data = UID()
    description = "This is a dummy id"
    tags = ["dummy", "test"]
    obj1 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([1, 2, 3, 4])
    description = "This is a dummy tensor n1"
    tags = ["dummy", "test"]
    obj2 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([10, 20, 30, 40])
    description = "This is a dummy tensor n2"
    tags = ["dummy", "test"]
    obj3 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = [obj1, obj2, obj3]
    description = "This is a dataset"
    tags = ["dummy", "dataset"]
    dataset_obj = Dataset(id=id, data=data, description=description, tags=tags)

    result = dataset_obj.__getitem__(_id=obj1.id)

    assert len(result) == 1
    assert result[0] == obj1


def test_dataset_get_element_fail() -> None:
    id = UID()
    data = UID()
    description = "This is a dummy id"
    tags = ["dummy", "test"]
    obj1 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([1, 2, 3, 4])
    description = "This is a dummy tensor n1"
    tags = ["dummy", "test"]
    obj2 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([10, 20, 30, 40])
    description = "This is a dummy tensor n2"
    tags = ["dummy", "test"]
    obj3 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = [obj1, obj2, obj3]
    description = "This is a dataset"
    tags = ["dummy", "dataset"]
    dataset_obj = Dataset(id=id, data=data, description=description, tags=tags)

    assert dataset_obj.__getitem__(UID()) == []


def test_dataset_get_keys() -> None:
    id = UID()
    data = UID()
    description = "This is a dummy id"
    tags = ["dummy", "test"]
    obj1 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([1, 2, 3, 4])
    description = "This is a dummy tensor n1"
    tags = ["dummy", "test"]
    obj2 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([10, 20, 30, 40])
    description = "This is a dummy tensor n2"
    tags = ["dummy", "test"]
    obj3 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = [obj1, obj2, obj3]
    description = "This is a dataset"
    tags = ["dummy", "dataset"]
    dataset_obj = Dataset(id=id, data=data, description=description, tags=tags)

    assert dataset_obj.keys() == [obj1.id, obj2.id, obj3.id]


def test_dataset_del() -> None:
    id = UID()
    data = UID()
    description = "This is a dummy id"
    tags = ["dummy", "test"]
    obj1 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([1, 2, 3, 4])
    description = "This is a dummy tensor n1"
    tags = ["dummy", "test"]
    obj2 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([10, 20, 30, 40])
    description = "This is a dummy tensor n2"
    tags = ["dummy", "test"]
    obj3 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = [obj1, obj2, obj3]
    description = "This is a dataset"
    tags = ["dummy", "dataset"]
    dataset_obj = Dataset(id=id, data=data, description=description, tags=tags)

    dataset_obj.__delitem__(obj2.id)

    assert dataset_obj.data == [obj1, obj3]


def test_serde_storable_obj() -> None:
    id = UID()
    data = th.Tensor([-1, -2, -3, -4])
    description = "This is a dummy id"
    tags = ["dummy", "test"]
    obj1 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([1, 2, 3, 4])
    description = "This is a dummy tensor n1"
    tags = ["dummy", "test"]
    obj2 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([10, 20, 30, 40])
    description = "This is a dummy tensor n2"
    tags = ["dummy", "test"]
    obj3 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = [obj1, obj2, obj3]
    description = "This is a dataset"
    tags = ["dummy", "dataset"]
    dataset_obj = Dataset(id=id, data=data, description=description, tags=tags)

    blob = sy.serialize(obj=dataset_obj)
    sy.deserialize(blob=blob)


def test_serde_storable_obj_2() -> None:
    id = UID()
    data = th.Tensor([-1, -2, -3, -4])
    description = "This is a dummy id"
    tags = ["dummy", "test"]
    obj1 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([1, 2, 3, 4])
    description = "This is a dummy tensor n1"
    tags = ["dummy", "test"]
    obj2 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = th.Tensor([10, 20, 30, 40])
    description = "This is a dummy tensor n2"
    tags = ["dummy", "test"]
    obj3 = StorableObject(id=id, data=data, description=description, tags=tags)

    id = UID()
    data = [obj1, obj2, obj3]
    description = "This is a dataset"
    tags = ["dummy", "dataset"]
    obj = Dataset(id=id, data=data, description=description, tags=tags)

    blob = serialize(obj)
    ds_obj = sy.deserialize(blob=blob)

    assert obj.id == ds_obj.id
    assert obj.description == ds_obj.description
    assert obj.tags == ds_obj.tags
    assert len(obj.data) == len(ds_obj.data)

    for i in range(len(obj.data)):
        assert obj.data[i].id == ds_obj.data[i].id
        assert th.all(th.eq(obj.data[i].data, ds_obj.data[i].data))
        assert obj.data[i].description == ds_obj.data[i].description
        assert obj.data[i].tags == ds_obj.data[i].tags



"""
PySyft Duet (WebRTC)

This class aims to implement the PySyft Duet concept by using WebRTC protocol as a
connection channel in order to allow two different users to establish a direct
connection with high-quality Real-time Communication using private addresses.

The most common example showing how it can be used is the notebook demo example:

Two different jupyter / collab notebooks in different machines using private addresses
behind routers, proxies and firewalls can connect using a full-duplex channel
to perform machine learning and data science tasks, working as a client
and server at the same time.

PS 1: You need a signaling server running somewhere.
If you don't know any public address running this service, or want to set up your own
signaling network you can use PyGrid's network app.

For local development you can run:
$ python src/syft/grid/example_nodes/network.py

PS 2: The PyGrid repo has a complimentary branch that matches the current PySyft release.
To use this feature you must use the correct PyGrid branch.
(https://github.com/OpenMined/PyGrid/)

You can get more details about all this process, in the syft/grid/connections/webrtc.py
source code.
"""

# stdlib
import asyncio
from typing import Optional

# third party
from nacl.signing import SigningKey

# syft relative
from ... import serialize
from ...core.io.route import SoloRoute
from ...core.node.common.metadata import Metadata
from ...core.node.domain.client import DomainClient
from ...core.node.domain.domain import Domain
from ...logger import error
from ...logger import traceback_and_raise
from ..connections.webrtc import WebRTCConnection
from ..duet.signaling_client import SignalingClient
from ..services.signaling_service import AnswerPullRequestMessage
from ..services.signaling_service import InvalidLoopBackRequest
from ..services.signaling_service import OfferPullRequestMessage
from ..services.signaling_service import SignalingAnswerMessage
from ..services.signaling_service import SignalingOfferMessage


class Duet(DomainClient):
    def __init__(
        self,
        node: Domain,
        target_id: str,
        signaling_client: SignalingClient,
        offer: bool = True,
    ):
        # Generate a signing key
        self.signing_key = SigningKey.generate()
        self.verify_key = self.signing_key.verify_key

        # Async Queues
        # These queues will be used in order to enqueue/dequeue
        # messages to be sent to the signaling server.
        self._push_msg_queue: asyncio.Queue = asyncio.Queue()
        self._pull_msg_queue: asyncio.Queue = asyncio.Queue()

        # As we need to inject a node instance inside of
        # a bidirectional connection in order to allow this
        # connection to work as a client and server using the
        # same channel. We need to be aware about forwarding
        # node instance references in order to avoid multiple
        # references to the same object (this makes the garbage
        # collecting difficult).
        # A good solution to avoid this problem is forward just
        # weak references. These references works like a proxy
        # not creating a  strong reference to the object.
        # So, If we delete the real object instance, the
        # garbage collect will call the __del__ method without problem.
        self.node = node

        # WebRTCConnection instance ( Bidirectional Connection )
        self.connection = WebRTCConnection(node=self.node)

        # Client used to exchange signaling messages in order to establish a connection
        # NOTE: In the future it may be a good idea to modularize this client to make
        # it pluggable using different connection protocols.
        self.signaling_client = signaling_client

        # If this peer will not start the signaling process
        if not offer:
            # Start adding an OfferPullRequest in order to verify if
            # the desired address pushed an offer request to connect with you.
            # This will trigger the pull async task to be check signaling notifications
            self._pull_msg_queue.put_nowait(
                OfferPullRequestMessage(
                    address=self.signaling_client.address,
                    target_peer=target_id,
                    host_peer=self.signaling_client.duet_id,
                    reply_to=self.signaling_client.address,
                )
            )
        else:
            # Push a WebRTC offer request to the address.
            self.send_offer(target_id=target_id)

        # This flag is used in order to finish the signaling process gracefully
        # While self._available is True, the pull/push tasks will be running
        # This flag will be setted to false when:
        # 1 - End of the signaling process (checked by _update_availability()).
        # 2 - Any Exception raised during these tasks.
        self._available = True

        # This attribute will be setted during the signaling messages exchange,
        # and used to create a SoloRoute for the both sides.
        self._client_metadata: Optional[Metadata] = None

        # Start async tasks and wait until one of them finishes.
        # As mentioned before, these tasks can be finished by two reasons:
        # 1 - Signaling process ends
        # 2 - Unexpected Exception
        try:
            asyncio.run(self.notify())

            # If client_metadata != None, then the connection was created successfully.
            if self._client_metadata is not None:
                # Deserialize client's metadata in order to obtain
                # PySyft's location structure
                (  # type: ignore
                    spec_location,
                    name,
                    _,
                ) = DomainClient.deserialize_client_metadata_from_node(
                    metadata=serialize(self._client_metadata)
                )

                # Create a SoloRoute
                route = SoloRoute(destination=spec_location, connection=self.connection)

                # Intialize the super class
                super().__init__(
                    domain=spec_location,
                    name=name,
                    routes=[route],
                    signing_key=self.signing_key,
                    verify_key=self.verify_key,
                )
                self.connection._client_address = self.address
            # If client_metada is None, then an exception occurred during the process
            # The exception has been caught and saved in self._exception
            else:
                # NOTE: Maybe we should create a custom exception type.
                traceback_and_raise(
                    Exception(
                        f"Something went wrong during the Duet init process. {self._exception}"
                    )
                )
        except Exception as e:
            traceback_and_raise(e)

    async def notify(self) -> None:
        try:
            # Enqueue Pull/Push async tasks
            push_task = asyncio.ensure_future(self.push())
            pull_task = asyncio.ensure_future(self.pull())

            # Wait until one of them finishes
            done, pending = await asyncio.wait(
                [pull_task, push_task], return_when=asyncio.FIRST_COMPLETED
            )

            # Finish the pending one.
            for task in pending:
                task.cancel()
        except Exception as e:
            traceback_and_raise(e)

    def close(self) -> None:
        self.connection.close()

    async def push(self) -> None:
        # This task is responsible for pushing offer/answer messages.
        try:
            while self._available:
                # If push_msg_queue is empty,
                # give up task queue priority, giving
                # computing time to the next task.
                msg = await self._push_msg_queue.get()
                # If self.push_msg_queue.get() returned a message (SignalingOfferMessage,SignalingAnswerMessage)
                # send it to the signaling server.
                self.signaling_client.send_immediate_msg_without_reply(msg=msg)
        except Exception as e:
            log = f"Got an exception in Duet push. {e}"
            error(log)
            # If any exception raises, set the self._available flag to False
            # in order to finish gracefully all the async tasks and save the exception.
            self._available = False
            self._exception: Exception = e

    async def pull(self) -> None:
        try:
            while self._available:
                # If pull_msg_queue is empty,
                # give up task queue priority, giving
                # computing time to the next task.
                msg = await self._pull_msg_queue.get()

                # If self.pull_msg_queue.get() returned a message (OfferPullRequestMessage,AnswerPullRequestMessage)
                # send it to the signaling server.
                _response = self.signaling_client.send_immediate_msg_with_reply(msg=msg)

                # If Signaling Offer Message was found
                if isinstance(_response, SignalingOfferMessage):
                    await self._send_answer(msg=_response)
                # If Signaling Answer Message was found
                elif isinstance(_response, SignalingAnswerMessage):
                    await self._ack(msg=_response)
                # If LoopBack Message it was a loopback request
                elif isinstance(_response, InvalidLoopBackRequest):
                    traceback_and_raise(
                        Exception(
                            "You can't perform p2p connection using your current node address as a destination peer."
                        )
                    )

                # If Signaling Message weren't found
                else:
                    # Just enqueue the request to be processed later.
                    self._pull_msg_queue.put_nowait(msg)

                # Checks if the signaling process is over.
                self._available = self._update_availability()
                await asyncio.sleep(0.5)

        except Exception as e:
            log = f"Got an exception in Duet pull. {e}"
            error(log)
            # If any exception raises, set the self._available flag to False
            # in order to finish gracefully all the async tasks and save the exception.
            self._available = False
            self._exception = e

    def send_offer(self, target_id: str) -> None:
        """Starts a new signaling process by creating a new
        offer message and pushing it to the Signaling Server."""
        try:
            # Generates an offer request payload containing
            # local network description data/metadata (IP, MAC, Mask, etc...)
            payload = asyncio.run(self.connection._set_offer())

            # Creates a PySyft's SignalingOfferMessage
            signaling_offer = SignalingOfferMessage(
                address=self.signaling_client.address,  # Target's address
                payload=payload,  # Offer Payload
                host_metadata=self.node.get_metadata_for_client(),  # Own Node Metadata
                target_peer=target_id,
                host_peer=self.signaling_client.duet_id,  # Own Node ID
            )

            # Enqueue it in push msg queue to be sent to the signaling server.
            self._push_msg_queue.put_nowait(signaling_offer)

            # Create/enqueue a new AnswerPullRequest in order to wait for signaling response.
            self._pull_msg_queue.put_nowait(
                AnswerPullRequestMessage(
                    address=self.signaling_client.address,
                    target_peer=target_id,
                    host_peer=self.signaling_client.duet_id,
                    reply_to=self.signaling_client.address,
                )
            )
        except Exception as e:
            traceback_and_raise(e)

    async def _send_answer(self, msg: SignalingOfferMessage) -> None:
        """Process SignalingOfferMessage and create a new
        SignalingAnswerMessage as a response"""
        try:
            # Process received offer message updating target's remote address
            # Generates an answer request payload containing
            # local network description data/metadata (IP, MAC, Mask, etc...)
            payload = asyncio.run(self.connection._set_answer(payload=msg.payload))

            # Save remote node's metadata in roder to create a SoloRoute.
            self._client_metadata = msg.host_metadata

            # Create a new SignalingAnswerMessage
            signaling_answer = SignalingAnswerMessage(
                address=self.signaling_client.address,
                payload=payload,  # Signaling answer payload
                host_metadata=self.node.get_metadata_for_client(),  # Own Node Metadata
                target_peer=msg.host_peer,  # Remote Node ID
                host_peer=self.signaling_client.duet_id,
            )

            # Enqueue it in the push msg queue to be sent to the signaling server.
            await self._push_msg_queue.put(signaling_answer)
        except Exception as e:
            traceback_and_raise(e)

    async def _ack(self, msg: SignalingAnswerMessage) -> None:
        """Last signaling message, stores remote Node
        metadata and updates target's remote address"""

        try:
            # Save remote node's metadata in roder to create a SoloRoute.
            self._client_metadata = msg.host_metadata

            # Process received offer message updating target's remote address
            await self.connection._process_answer(payload=msg.payload)
        except Exception as e:
            traceback_and_raise(e)

    def _update_availability(self) -> bool:
        """Method used to check if the signaling process is over.
        :return: Boolean flag, True if it's NOT over, and False if it's over.
        :rtype: Boolean
        """
        try:
            return (
                not self._pull_msg_queue.empty()
                and self.connection.peer_connection is not None
            )
        except Exception as e:
            traceback_and_raise(e)



# DOs and Don's of this class:
# - Do NOT use absolute syft imports (i.e. import syft.core...) Use relative ones.
# - Do NOT put multiple imports on the same line (i.e. from <x> import a, b, c). Use separate lines
# - Do sort imports by length
# - Do group imports by where they come from

# stdlib
from typing import List
from typing import Optional
from typing import Type

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey
from typing_extensions import final

# syft relative
from ..... import serialize
from .....core.common.serde.serializable import bind_protobuf
from .....logger import error
from .....proto.core.node.common.service.object_search_message_pb2 import (
    ObjectSearchMessage as ObjectSearchMessage_PB,
)
from .....proto.core.node.common.service.object_search_message_pb2 import (
    ObjectSearchReplyMessage as ObjectSearchReplyMessage_PB,
)
from .....util import obj2pointer_type
from .....util import traceback_and_raise
from ....common.group import VERIFYALL
from ....common.message import ImmediateSyftMessageWithReply
from ....common.message import ImmediateSyftMessageWithoutReply
from ....common.serde.deserialize import _deserialize
from ....common.uid import UID
from ....io.address import Address
from ....pointer.pointer import Pointer
from ...abstract.node import AbstractNode
from .node_service import ImmediateNodeServiceWithReply


@bind_protobuf
@final
class ObjectSearchMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self, address: Address, reply_to: Address, msg_id: Optional[UID] = None
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        """By default this message just returns pointers to all the objects
        the sender is allowed to see. In the future we'll add support so that
        we can query for subsets."""

    def _object2proto(self) -> ObjectSearchMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: ObjectSearchMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return ObjectSearchMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(proto: ObjectSearchMessage_PB) -> "ObjectSearchMessage":
        """Creates a ObjectSearchMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of ObjectSearchMessage
        :rtype: ObjectSearchMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return ObjectSearchMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return ObjectSearchMessage_PB


@bind_protobuf
@final
class ObjectSearchReplyMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        results: List[Pointer],
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        """By default this message just returns pointers to all the objects
        the sender is allowed to see. In the future we'll add support so that
        we can query for subsets."""
        self.results = results

    def _object2proto(self) -> ObjectSearchReplyMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: ObjectSearchReplyMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return ObjectSearchReplyMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            results=list(map(lambda x: serialize(x), self.results)),
        )

    @staticmethod
    def _proto2object(proto: ObjectSearchReplyMessage_PB) -> "ObjectSearchReplyMessage":
        """Creates a ObjectSearchReplyMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of ObjectSearchReplyMessage
        :rtype: ObjectSearchReplyMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return ObjectSearchReplyMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            results=[_deserialize(blob=x) for x in proto.results],
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return ObjectSearchReplyMessage_PB


class ImmediateObjectSearchService(ImmediateNodeServiceWithReply):
    @staticmethod
    def process(
        node: AbstractNode,
        msg: ObjectSearchMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> ObjectSearchReplyMessage:
        results: List[Pointer] = list()

        if verify_key is None:
            traceback_and_raise(
                "Can't process an ImmediateObjectSearchService with no "
                "verification key."
            )

        try:
            for obj in node.store.get_objects_of_type(obj_type=object):
                # if this tensor allows anyone to search for it, then one of its keys
                # has a VERIFYALL in it.
                contains_all_in_permissions = any(
                    key is VERIFYALL for key in obj.search_permissions.keys()
                )
                if (
                    verify_key in obj.search_permissions.keys()
                    or verify_key == node.root_verify_key
                    or contains_all_in_permissions
                ):
                    ptr_type = obj2pointer_type(obj=obj.data)
                    ptr = ptr_type(
                        client=node,
                        id_at_location=obj.id,
                        object_type=obj.object_type,
                        tags=obj.tags,
                        description=obj.description,
                    )
                    results.append(ptr)
        except Exception as e:
            error(f"Error searching store. {e}")

        return ObjectSearchReplyMessage(address=msg.reply_to, results=results)

    @staticmethod
    def message_handler_types() -> List[Type[ObjectSearchMessage]]:
        return [ObjectSearchMessage]



# stdlib
from typing import Any


# Decryption shouldn't be possible over Duet pointers.
# We retrieve the data, link it to the secret context, and then decrypt.
def decrypt(secret_ctx: Any, ptr: Any) -> Any:
    local = ptr.get()
    local.link_context(secret_ctx)

    return local.decrypt()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/action/run_class_method.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.core.pointer import (
    pointer_pb2 as proto_dot_core_dot_pointer_dot_pointer__pb2,
)

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/core/node/common/action/run_class_method.proto",
    package="syft.core.node.common.action",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n4proto/core/node/common/action/run_class_method.proto\x12\x1csyft.core.node.common.action\x1a%proto/core/common/common_object.proto\x1a proto/core/pointer/pointer.proto\x1a\x1bproto/core/io/address.proto"\x92\x03\n\x14RunClassMethodAction\x12\x0c\n\x04path\x18\x01 \x01(\t\x12)\n\x05_self\x18\x02 \x01(\x0b\x32\x1a.syft.core.pointer.Pointer\x12(\n\x04\x61rgs\x18\x03 \x03(\x0b\x32\x1a.syft.core.pointer.Pointer\x12N\n\x06kwargs\x18\x04 \x03(\x0b\x32>.syft.core.node.common.action.RunClassMethodAction.KwargsEntry\x12-\n\x0eid_at_location\x18\x05 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x06 \x01(\x0b\x32\x15.syft.core.io.Address\x12%\n\x06msg_id\x18\x07 \x01(\x0b\x32\x15.syft.core.common.UID\x1aI\n\x0bKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12)\n\x05value\x18\x02 \x01(\x0b\x32\x1a.syft.core.pointer.Pointer:\x02\x38\x01\x62\x06proto3',
    dependencies=[
        proto_dot_core_dot_common_dot_common__object__pb2.DESCRIPTOR,
        proto_dot_core_dot_pointer_dot_pointer__pb2.DESCRIPTOR,
        proto_dot_core_dot_io_dot_address__pb2.DESCRIPTOR,
    ],
)


_RUNCLASSMETHODACTION_KWARGSENTRY = _descriptor.Descriptor(
    name="KwargsEntry",
    full_name="syft.core.node.common.action.RunClassMethodAction.KwargsEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="syft.core.node.common.action.RunClassMethodAction.KwargsEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="syft.core.node.common.action.RunClassMethodAction.KwargsEntry.value",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=518,
    serialized_end=591,
)

_RUNCLASSMETHODACTION = _descriptor.Descriptor(
    name="RunClassMethodAction",
    full_name="syft.core.node.common.action.RunClassMethodAction",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="path",
            full_name="syft.core.node.common.action.RunClassMethodAction.path",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="_self",
            full_name="syft.core.node.common.action.RunClassMethodAction._self",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="args",
            full_name="syft.core.node.common.action.RunClassMethodAction.args",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="kwargs",
            full_name="syft.core.node.common.action.RunClassMethodAction.kwargs",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="id_at_location",
            full_name="syft.core.node.common.action.RunClassMethodAction.id_at_location",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.core.node.common.action.RunClassMethodAction.address",
            index=5,
            number=6,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.core.node.common.action.RunClassMethodAction.msg_id",
            index=6,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[
        _RUNCLASSMETHODACTION_KWARGSENTRY,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=189,
    serialized_end=591,
)

_RUNCLASSMETHODACTION_KWARGSENTRY.fields_by_name[
    "value"
].message_type = proto_dot_core_dot_pointer_dot_pointer__pb2._POINTER
_RUNCLASSMETHODACTION_KWARGSENTRY.containing_type = _RUNCLASSMETHODACTION
_RUNCLASSMETHODACTION.fields_by_name[
    "_self"
].message_type = proto_dot_core_dot_pointer_dot_pointer__pb2._POINTER
_RUNCLASSMETHODACTION.fields_by_name[
    "args"
].message_type = proto_dot_core_dot_pointer_dot_pointer__pb2._POINTER
_RUNCLASSMETHODACTION.fields_by_name[
    "kwargs"
].message_type = _RUNCLASSMETHODACTION_KWARGSENTRY
_RUNCLASSMETHODACTION.fields_by_name[
    "id_at_location"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_RUNCLASSMETHODACTION.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_RUNCLASSMETHODACTION.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
DESCRIPTOR.message_types_by_name["RunClassMethodAction"] = _RUNCLASSMETHODACTION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RunClassMethodAction = _reflection.GeneratedProtocolMessageType(
    "RunClassMethodAction",
    (_message.Message,),
    {
        "KwargsEntry": _reflection.GeneratedProtocolMessageType(
            "KwargsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _RUNCLASSMETHODACTION_KWARGSENTRY,
                "__module__": "proto.core.node.common.action.run_class_method_pb2"
                # @@protoc_insertion_point(class_scope:syft.core.node.common.action.RunClassMethodAction.KwargsEntry)
            },
        ),
        "DESCRIPTOR": _RUNCLASSMETHODACTION,
        "__module__": "proto.core.node.common.action.run_class_method_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.action.RunClassMethodAction)
    },
)
_sym_db.RegisterMessage(RunClassMethodAction)
_sym_db.RegisterMessage(RunClassMethodAction.KwargsEntry)


_RUNCLASSMETHODACTION_KWARGSENTRY._options = None
# @@protoc_insertion_point(module_scope)



# syft absolute
from syft.lib.python.primitive_factory import PrimitiveFactory
from syft.lib.python.util import downcast_args_and_kwargs


def test_downcast():
    assert downcast_args_and_kwargs(args=[1, 2, 3], kwargs={}) == ([1, 2, 3], {})
    assert downcast_args_and_kwargs(args=["test", True], kwargs={1: 2}) == (
        ["test", True],
        {1: 2},
    )


def test_recurse():
    assert PrimitiveFactory.generate_primitive(value={1: 1, 2: 2}, recurse=True) == {
        1: 1,
        2: 2,
    }



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/plan/plan.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.node.common.action import (
    action_pb2 as proto_dot_core_dot_node_dot_common_dot_action_dot_action__pb2,
)
from syft.proto.core.pointer import (
    pointer_pb2 as proto_dot_core_dot_pointer_dot_pointer__pb2,
)

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/core/plan/plan.proto",
    package="syft.core.plan",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x1aproto/core/plan/plan.proto\x12\x0esyft.core.plan\x1a*proto/core/node/common/action/action.proto\x1a proto/core/pointer/pointer.proto"\xe7\x01\n\x04Plan\x12\x35\n\x07\x61\x63tions\x18\x01 \x03(\x0b\x32$.syft.core.node.common.action.Action\x12\x30\n\x06inputs\x18\x02 \x03(\x0b\x32 .syft.core.plan.Plan.InputsEntry\x12+\n\x07outputs\x18\x03 \x03(\x0b\x32\x1a.syft.core.pointer.Pointer\x1aI\n\x0bInputsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12)\n\x05value\x18\x02 \x01(\x0b\x32\x1a.syft.core.pointer.Pointer:\x02\x38\x01\x62\x06proto3',
    dependencies=[
        proto_dot_core_dot_node_dot_common_dot_action_dot_action__pb2.DESCRIPTOR,
        proto_dot_core_dot_pointer_dot_pointer__pb2.DESCRIPTOR,
    ],
)


_PLAN_INPUTSENTRY = _descriptor.Descriptor(
    name="InputsEntry",
    full_name="syft.core.plan.Plan.InputsEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="syft.core.plan.Plan.InputsEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="syft.core.plan.Plan.InputsEntry.value",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=283,
    serialized_end=356,
)

_PLAN = _descriptor.Descriptor(
    name="Plan",
    full_name="syft.core.plan.Plan",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="actions",
            full_name="syft.core.plan.Plan.actions",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="inputs",
            full_name="syft.core.plan.Plan.inputs",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="outputs",
            full_name="syft.core.plan.Plan.outputs",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[
        _PLAN_INPUTSENTRY,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=125,
    serialized_end=356,
)

_PLAN_INPUTSENTRY.fields_by_name[
    "value"
].message_type = proto_dot_core_dot_pointer_dot_pointer__pb2._POINTER
_PLAN_INPUTSENTRY.containing_type = _PLAN
_PLAN.fields_by_name[
    "actions"
].message_type = proto_dot_core_dot_node_dot_common_dot_action_dot_action__pb2._ACTION
_PLAN.fields_by_name["inputs"].message_type = _PLAN_INPUTSENTRY
_PLAN.fields_by_name[
    "outputs"
].message_type = proto_dot_core_dot_pointer_dot_pointer__pb2._POINTER
DESCRIPTOR.message_types_by_name["Plan"] = _PLAN
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Plan = _reflection.GeneratedProtocolMessageType(
    "Plan",
    (_message.Message,),
    {
        "InputsEntry": _reflection.GeneratedProtocolMessageType(
            "InputsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _PLAN_INPUTSENTRY,
                "__module__": "proto.core.plan.plan_pb2"
                # @@protoc_insertion_point(class_scope:syft.core.plan.Plan.InputsEntry)
            },
        ),
        "DESCRIPTOR": _PLAN,
        "__module__": "proto.core.plan.plan_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.plan.Plan)
    },
)
_sym_db.RegisterMessage(Plan)
_sym_db.RegisterMessage(Plan.InputsEntry)


_PLAN_INPUTSENTRY._options = None
# @@protoc_insertion_point(module_scope)
